import asyncio
import json
import logging
import random
import time
from asyncio import Task
from collections import deque
from contextlib import asynccontextmanager
from datetime import datetime, timezone
from json import JSONDecodeError
from typing import AsyncGenerator, Any, Tuple

import numpy as np
import psutil
import zmq.asyncio
from fastapi import FastAPI
from psutil._common import snicaddr, snetio, snicstats
from pydantic import BaseModel, Field, field_serializer
from pydantic_core import Url

from dranspose.helpers.utils import cancel_and_wait, done_callback
from tests.stream1 import AcquisitionSocket

logger = logging.getLogger(__name__)


class SocketSpec(BaseModel):
    type: int = zmq.PUSH
    port: int = 9999


class ConnectSpec(BaseModel):
    type: int = zmq.PULL
    url: Url = Url("tcp://127.0.0.1:9999")


class WorkloadSpec(BaseModel):
    number: int = 1
    time: float = 0.1
    shape: Tuple[int, int] = (10, 10)
    filename: str = ""


class Statistics(BaseModel):
    snapshots: deque[tuple[float, int, dict[str, snetio]]] = Field(
        default_factory=lambda: deque(maxlen=100)
    )
    fps: float = 0
    packets: int = 0
    measured: float = 0
    deltas: list[float] = []

    @field_serializer("snapshots", mode="wrap")
    def serialize_sn(
        self, snapshots: deque[tuple[float, int, dict[str, snetio]]], _info: Any
    ) -> list[tuple[float, int, dict[str, snetio]]]:
        return list(snapshots)


class NetworkConfig(BaseModel):
    addresses: dict[str, list[snicaddr]]
    stats: dict[str, snicstats]


class WorkloadGenerator:
    def __init__(self, **kwargs: Any) -> None:
        self.context = zmq.asyncio.Context()
        self.out_socket: AcquisitionSocket | None = None
        self.in_socket: zmq.asyncio.Socket | None = None
        self.task: Task[Any] | None = None
        self.stat = Statistics()
        logger.info("Workload generator initialised")

    def finished(self) -> bool:
        if self.task is None:
            return True
        return self.task.done()

    async def calc_stat(self) -> None:
        start = time.time()
        before_packets = self.stat.packets
        while True:
            await asyncio.sleep(0.5)
            after = time.time()
            num = self.stat.packets
            ctr = psutil.net_io_counters(pernic=True)
            self.stat.measured = after
            self.stat.fps = (num - before_packets) / (after - start)
            self.stat.snapshots.append((after, num, ctr))
            logger.debug("calculated stats %s fps", self.stat.fps)
            start = after
            before_packets = num

    async def open_socket(self, spec: SocketSpec) -> None:
        await self.close_socket()
        self.stat = Statistics()
        self.out_socket = AcquisitionSocket(
            self.context, Url(f"tcp://*:{spec.port}"), typ=spec.type
        )
        logger.info("socket opened to %s is %s", spec, self.out_socket)

    async def sink_packets(self) -> None:
        if not isinstance(self.in_socket, zmq.Socket):
            raise Exception("only sink packets from connected socket")
        while True:
            pkt = await self.in_socket.recv_multipart(copy=False)
            try:
                header = json.loads(pkt[0].bytes)
                ts = header.get("timestamps", {})
                now = datetime.now(timezone.utc)
                delta = now - min(map(lambda x: datetime.fromisoformat(x), ts.values()))
                logger.debug("delta is %s", delta.total_seconds())
                self.stat.deltas.append(delta.total_seconds())
            except (JSONDecodeError, ValueError):
                pass
            logger.debug("sinked packet %s", pkt)
            self.stat.packets += 1

    async def connect_socket(self, spec: ConnectSpec) -> None:
        await self.close_socket()
        self.stat = Statistics()
        self.in_socket = self.context.socket(spec.type)
        self.in_socket.connect(str(spec.url))
        if spec.type == zmq.SUB:
            self.in_socket.setsockopt(zmq.SUBSCRIBE, b"")
        logger.info("socket connected to %s", spec)
        self.task = asyncio.create_task(self.sink_packets())
        self.task.add_done_callback(done_callback)

    async def send_packets(self, spec: WorkloadSpec) -> None:
        logger.info("send packets from socket %s", self.out_socket)
        if self.out_socket is None:
            raise Exception("must open socket before sending packets")
        if not isinstance(self.out_socket, AcquisitionSocket):
            raise Exception("must be AcquisitionSocket to send, is %s", self.out_socket)
        acq = await self.out_socket.start(filename=spec.filename)
        logger.info("sending packets %s", spec)
        self.stat.packets += 1
        width = spec.shape[0]
        height = spec.shape[1]
        img = np.zeros((width, height), dtype=np.uint16)
        for _ in range(20):
            img[random.randint(0, width - 1)][
                random.randint(0, height - 1)
            ] = random.randint(0, 10)
        for frameno in range(spec.number):
            extra = {
                "timestamps": {"generator": datetime.now(timezone.utc).isoformat()}
            }
            await acq.image(img, img.shape, frameno, extra_fields=extra)
            logger.debug("sent frame %d", frameno)
            self.stat.packets += 1
            await asyncio.sleep(spec.time)
        await acq.close()
        self.stat.packets += 1

    async def close_socket(self) -> None:
        if self.in_socket is not None:
            self.in_socket.close()
        if self.out_socket is not None:
            await self.out_socket.close()
        logger.info("socket closed")

    async def close(self) -> None:
        self.context.destroy(linger=0)
        logger.info("context destroyed")


gen: WorkloadGenerator


@asynccontextmanager
async def lifespan(app: FastAPI) -> AsyncGenerator[None, None]:
    # Load the ML model
    global gen
    gen = WorkloadGenerator()
    task = asyncio.create_task(gen.calc_stat())
    yield
    await cancel_and_wait(task)
    await gen.close()
    # Clean up the ML models and release the resources


app = FastAPI(lifespan=lifespan)


@app.get("/api/v1/finished")
async def get_fin() -> bool:
    global gen
    return gen.finished()


@app.get("/api/v1/config")
async def get_conf() -> NetworkConfig:
    return NetworkConfig(addresses=psutil.net_if_addrs(), stats=psutil.net_if_stats())


@app.get("/api/v1/statistics")
async def get_stat() -> Statistics:
    global gen
    return gen.stat


@app.post("/api/v1/open_socket")
async def open_sock(sockspec: SocketSpec) -> SocketSpec:
    global gen
    print(sockspec)
    await gen.open_socket(sockspec)
    return sockspec


@app.post("/api/v1/frames")
async def frames(spec: WorkloadSpec) -> bool:
    global gen
    gen.task = asyncio.create_task(gen.send_packets(spec))
    gen.task.add_done_callback(done_callback)
    return True


@app.post("/api/v1/close_socket")
async def close_sock() -> bool:
    global gen
    await gen.close_socket()
    return True


@app.post("/api/v1/connect_socket")
async def connect_sock(sockspec: ConnectSpec) -> ConnectSpec:
    global gen
    await gen.connect_socket(sockspec)
    return sockspec



import asyncio
import json
import logging
import pickle
import traceback
from contextlib import asynccontextmanager
from typing import Optional, AsyncGenerator, Any

import zmq.asyncio
from fastapi import FastAPI, HTTPException
from pydantic import UUID4
from starlette.responses import Response

from dranspose.helpers import utils
from dranspose.distributed import DistributedService, DistributedSettings
from dranspose.event import ResultData
from dranspose.helpers.h5dict import router
from dranspose.helpers.jsonpath_slice_ext import NumpyExtentedJsonPathParser
from dranspose.helpers.utils import done_callback, cancel_and_wait
from dranspose.protocol import (
    ReducerState,
    ZmqUrl,
    RedisKeys,
    ReducerUpdate,
    DistributedStateEnum,
    StreamName,
)

logger = logging.getLogger(__name__)


class ReducerSettings(DistributedSettings):
    reducer_url: ZmqUrl = ZmqUrl("tcp://localhost:10200")
    reducer_class: Optional[str] = None


class Reducer(DistributedService):
    def __init__(self, settings: Optional[ReducerSettings] = None):
        self._reducer_settings = settings
        if self._reducer_settings is None:
            self._reducer_settings = ReducerSettings()

        state = ReducerState(url=self._reducer_settings.reducer_url)
        super().__init__(state, self._reducer_settings)
        self._logger.info("created reducer with state %s", state)
        self.state: ReducerState
        self.ctx = zmq.asyncio.Context()
        self.in_socket = self.ctx.socket(zmq.PULL)
        self.in_socket.setsockopt(zmq.TCP_KEEPALIVE, 1)
        self.in_socket.setsockopt(zmq.TCP_KEEPALIVE_IDLE, 300)
        self.in_socket.setsockopt(zmq.TCP_KEEPALIVE_INTVL, 300)
        self.in_socket.bind(f"tcp://*:{self._reducer_settings.reducer_url.port}")

        self.custom = None
        self.custom_context: dict[Any, Any] = {}
        if self._reducer_settings.reducer_class:
            try:
                self.custom = utils.import_class(self._reducer_settings.reducer_class)
                self._logger.info("custom reducer class %s", self.custom)
                try:
                    self.param_descriptions = self.custom.describe_parameters()  # type: ignore[attr-defined]
                except AttributeError:
                    self._logger.info(
                        "custom worker class has no describe_parameters staticmethod"
                    )
                except Exception as e:
                    self._logger.error(
                        "custom worker parameter descripition is broken: %s",
                        e.__repr__(),
                    )
            except Exception as e:
                self._logger.warning(
                    "failed to load custom reducer class, discarding results %s trace: %s",
                    e.__repr__(),
                    traceback.format_exc(),
                )

    async def run(self) -> None:
        self.work_task = asyncio.create_task(self.work())
        self.work_task.add_done_callback(done_callback)
        self.timer_task = asyncio.create_task(self.timer())
        self.timer_task.add_done_callback(done_callback)
        self.metrics_task = asyncio.create_task(self.update_metrics())
        self.metrics_task.add_done_callback(done_callback)
        await self.register()

    async def work(self) -> None:
        self._logger.info("started work task")
        self.reducer = None
        if self.custom:
            self.reducer = self.custom(
                parameters=self.parameters,
                context=self.custom_context,
                state=self.state,
            )
        while True:
            parts = await self.in_socket.recv_multipart()
            prelim = json.loads(parts[0])
            prelim["payload"] = pickle.loads(parts[1])
            result = ResultData.model_validate(prelim)
            if self.reducer:
                try:
                    loop = asyncio.get_event_loop()
                    await loop.run_in_executor(
                        None, self.reducer.process_result, result, self.parameters
                    )
                except Exception as e:
                    self._logger.error(
                        "custom reducer failed: %s\n%s",
                        e.__repr__(),
                        traceback.format_exc(),
                    )
            ru = ReducerUpdate(
                state=DistributedStateEnum.IDLE,
                completed=result.event_number,
                worker=result.worker,
            )
            self._logger.debug("result processed, notify controller with %s", ru)
            await self.redis.xadd(
                RedisKeys.ready(self.state.mapping_uuid),
                {"data": ru.model_dump_json()},
            )
            self._logger.debug("processed result %s", result)
            self.state.processed_events += 1

    async def timer(self) -> None:
        self._logger.info("started timer task")
        while True:
            delay = 1
            if self.reducer:
                if hasattr(self.reducer, "timer"):
                    try:
                        loop = asyncio.get_event_loop()
                        delay = await loop.run_in_executor(None, self.reducer.timer)
                    except Exception as e:
                        self._logger.error(
                            "custom reducer timer failed: %s\n%s",
                            e.__repr__(),
                            traceback.format_exc(),
                        )

            if not isinstance(delay, (int, float)):
                self._logger.error(
                    "custom reducer time is not a number: %s",
                    delay,
                )
                delay = 1
            await asyncio.sleep(delay)

    async def restart_work(
        self, new_uuid: UUID4, active_streams: list[StreamName]
    ) -> None:
        self._logger.info("resetting config %s", new_uuid)
        await cancel_and_wait(self.timer_task)
        await cancel_and_wait(self.work_task)
        self.state.mapping_uuid = new_uuid
        self.work_task = asyncio.create_task(self.work())
        self.work_task.add_done_callback(done_callback)
        self.timer_task = asyncio.create_task(self.timer())
        self.timer_task.add_done_callback(done_callback)

    async def finish_work(self) -> None:
        self._logger.info("finishing reducer work")
        if self.reducer:
            if hasattr(self.reducer, "finish"):
                try:
                    loop = asyncio.get_event_loop()
                    await loop.run_in_executor(
                        None, self.reducer.finish, self.parameters
                    )
                except Exception as e:
                    self._logger.error(
                        "custom reducer finish failed: %s\n%s",
                        e.__repr__(),
                        traceback.format_exc(),
                    )
        await self.redis.xadd(
            RedisKeys.ready(self.state.mapping_uuid),
            {
                "data": ReducerUpdate(
                    state=DistributedStateEnum.FINISHED,
                ).model_dump_json()
            },
        )

    async def close(self) -> None:
        if self.reducer:
            if hasattr(self.reducer, "close"):
                try:
                    loop = asyncio.get_event_loop()
                    await loop.run_in_executor(
                        None, self.reducer.close, self.custom_context
                    )
                except Exception as e:
                    self._logger.error(
                        "custom reducer failed to close: %s\n%s",
                        e.__repr__(),
                        traceback.format_exc(),
                    )
        await cancel_and_wait(self.timer_task)
        await cancel_and_wait(self.work_task)
        await cancel_and_wait(self.metrics_task)
        await self.redis.delete(RedisKeys.config("reducer", self.state.name))
        await super().close()
        self.ctx.destroy(linger=0)
        self._logger.info("closed reducer")


reducer: Reducer


@asynccontextmanager
async def lifespan(app: FastAPI) -> AsyncGenerator[None, None]:
    # Load the ML model
    global reducer
    reducer = Reducer()
    run_task = asyncio.create_task(reducer.run())
    run_task.add_done_callback(done_callback)

    def get_data() -> dict[str, Any]:
        if reducer.reducer is not None:
            if hasattr(reducer.reducer, "publish"):
                return reducer.reducer.publish
        return {}

    app.state.get_data = get_data
    yield
    await cancel_and_wait(run_task)
    await reducer.close()
    # Clean up the ML models and release the resources


app = FastAPI(lifespan=lifespan)

app.include_router(router)


@app.get("/api/v1/status")
async def get_status() -> bool:
    return True


@app.get("/api/v1/result/pickle")
async def get_result() -> Any | bytes:
    global reducer
    data = b""
    try:
        if hasattr(reducer.reducer, "publish"):
            data = pickle.dumps(reducer.reducer.publish)  # type: ignore [union-attr]
    except pickle.PicklingError:
        logging.warning("publishable data cannot be pickled")
    return Response(data, media_type="application/x.pickle")


@app.get("/api/v1/result/{path:path}")
async def get_path(path: str) -> Any:
    global reducer
    if not hasattr(reducer.reducer, "publish"):
        raise HTTPException(status_code=404, detail="no publishable data")
    try:
        if path == "":
            path = "$"
        jsonpath_expr = NumpyExtentedJsonPathParser(debug=False).parse(path)
        print("expr", jsonpath_expr.__repr__())
        ret = [match.value for match in jsonpath_expr.find(reducer.reducer.publish)]  # type: ignore [union-attr]
        data = pickle.dumps(ret)
        return Response(data, media_type="application/x.pickle")
    except Exception as e:
        raise HTTPException(status_code=400, detail="malformed path %s" % e.__repr__())



from typing import Awaitable, Callable, Optional
import zmq.asyncio

import aiohttp
import pytest
from pydantic_core import Url

from dranspose.ingester import Ingester
from dranspose.ingesters.zmqpull_single import (
    ZmqPullSingleIngester,
    ZmqPullSingleSettings,
)
from dranspose.protocol import (
    StreamName,
    WorkerName,
    VirtualWorker,
    VirtualConstraint,
)
from dranspose.worker import Worker
from tests.utils import wait_for_controller, wait_for_finish


@pytest.mark.asyncio
async def test_stop_scans(
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[WorkerName], Awaitable[Worker]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
) -> None:
    await reducer("examples.dummy.reducer:FluorescenceReducer")
    await create_worker(WorkerName("w1"))
    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("eiger")],
                upstream_url=Url("tcp://localhost:9999"),
            ),
        )
    )

    await wait_for_controller(streams={StreamName("eiger")})
    async with aiohttp.ClientSession() as session:
        ntrig = 30
        mp = {
            "eiger": [
                [VirtualWorker(constraint=VirtualConstraint(i)).model_dump(mode="json")]
                for i in range(ntrig)
            ],
        }
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json=mp,
        )
        assert resp.status == 200
        await resp.json()

    context = zmq.asyncio.Context()

    # asyncio.create_task()

    async with aiohttp.ClientSession() as session:
        resp = await session.post(
            "http://localhost:5000/api/v1/stop",
        )

    await wait_for_finish()

    context.destroy()



from typing import Optional, Any

from jsonpath_ng.ext.parser import ExtentedJsonPathParser  # type: ignore
from jsonpath_ng import DatumInContext, This  # type: ignore


class DefintionInvalid(Exception):
    pass


class Numpy(This):
    """Numpy slicer

    Concrete syntax is '`slice(numpy slice, multidim)`'
    """

    def _parse_slice(self, slice_str: str) -> Any | int | slice:
        if slice_str == "...":
            return Ellipsis
        elif slice_str.isdigit():
            return int(slice_str)
        else:
            return slice(*map(lambda x: int(x) if x else None, slice_str.split(":")))

    def __init__(self, method: Optional[str] = None) -> None:
        if not method:
            raise DefintionInvalid("method must be provided")
        method = method[len("numpy(") : -1]
        slice_objects = [self._parse_slice(s.strip()) for s in method.split(",")]
        if slice_objects is []:
            raise DefintionInvalid("%s is not valid" % method)
        self.slice = tuple(slice_objects)
        self.method = method

    def find(self, datum: DatumInContext) -> list[DatumInContext]:
        # print("datum got", datum)
        # datum = DatumInContext.wrap(datum)
        try:
            # value = datum.value[self.slice] #.split(self.char, self.max_split)[self.segment]
            datum = DatumInContext(datum.value[self.slice], path=self, context=datum)
        except Exception:
            return []
        return [datum]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Numpy) and self.method == other.method

    def __repr__(self) -> str:
        return "%s(%r)" % (self.__class__.__name__, self.method)

    def __str__(self) -> str:
        return "`%s`" % self.method


class NumpyExtentedJsonPathParser(ExtentedJsonPathParser):
    """Custom LALR-parser for JsonPath"""

    def p_jsonpath_named_operator(self, p: list[Any]) -> None:
        "jsonpath : NAMED_OPERATOR"
        if p[1].startswith("numpy("):
            p[0] = Numpy(p[1])
        else:
            super(NumpyExtentedJsonPathParser, self).p_jsonpath_named_operator(p)



from typing import Literal

from pydantic import BaseModel, TypeAdapter, ConfigDict


class AlbaemBase(BaseModel):
    message_id: int
    version: Literal[1]


class AlbaemStart(AlbaemBase):
    """
    Example:
        ``` py
        AlbaemStart(
            message_id=1,
            version=1,
            message_type='series-start'
        )
        ```
    """

    model_config = ConfigDict(extra="allow")

    message_type: Literal["series-start"]


class AlbaemData(AlbaemBase):
    """
    While the original stream sends 3 separate zmq frames (no multipart), this returns a single packet.

    Example:
        ``` py
        AlbaemImage(
            message_id=2,
            version=1,
            message_type='data',
            frame_number=0,
            timestamp=7.72645192,
            acquisition_timestamp=1704807770443944912,
            channel1=-2.8302592615927417e-11,
            channel2=-6.091210149949596e-11,
            channel3=2.5349278603830644e-10,
            channel4=4.80528800718246e-10
        )
        ```
    """

    model_config = ConfigDict(extra="allow")

    message_type: Literal["data"]
    frame_number: int
    timestamp: float
    acquisition_timestamp: int


class AlbaemEnd(AlbaemBase):
    """
    Example:
        ``` py
        AlbaemEnd(
            message_id=6,
            version=1,
            message_type='series-end',
            detector_specific={'read_overflow': False, 'memory_overflow': False})
        ```
    """

    model_config = ConfigDict(extra="allow")

    message_type: Literal["series-end"]


AlbaemPacket = TypeAdapter(AlbaemStart | AlbaemData | AlbaemEnd)
"""
Union type for Albaem packets
"""



import asyncio
import logging
import os
from pathlib import PosixPath
from typing import Awaitable, Callable, Coroutine, Optional, Any

import aiohttp
import h5pyd
import zmq.asyncio

import pytest
from pydantic_core import Url

from dranspose.ingester import Ingester
from dranspose.ingesters.zmqsub_lecroy import ZmqSubLecroyIngester, ZmqSubLecroySettings
from dranspose.protocol import (
    EnsembleState,
    StreamName,
    WorkerName,
    VirtualWorker,
    VirtualConstraint,
)

from dranspose.worker import Worker, WorkerSettings


@pytest.mark.parametrize(
    "filename, ntrig, nburst, max0, ts0",
    [
        (
            "tests/data/maui-02-continuous-3.cbor.gz",
            3,
            1,
            0.1675630532008654,
            1740563664.335,
        ),
        (
            "tests/data/maui-02-sequential-3.cbor.gz",
            3,
            20,
            0.022609806155742262,
            1740736919.435,
        ),
    ],
)
@pytest.mark.asyncio
async def test_lecroy(
    filename: str,
    ntrig: int,
    nburst: int,
    max0: float,
    ts0: float,
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[Worker], Awaitable[Worker]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    stream_pkls: Callable[
        [zmq.Context[Any], int, os.PathLike[Any] | str, float, int],
        Coroutine[Any, Any, None],
    ],
) -> None:
    await reducer("examples.parser.lecroy:LecroyReducer")
    await create_worker(
        Worker(
            settings=WorkerSettings(
                worker_name=WorkerName("w1"),
                worker_class="examples.parser.lecroy:LecroyWorker",
            ),
        )
    )
    await create_ingester(
        ZmqSubLecroyIngester(
            settings=ZmqSubLecroySettings(
                ingester_streams=[StreamName("lecroy")],
                upstream_url=Url("tcp://localhost:22004"),
            ),
        )
    )

    async with aiohttp.ClientSession() as session:
        st = await session.get("http://localhost:5000/api/v1/config")
        state = EnsembleState.model_validate(await st.json())
        while {"lecroy"} - set(state.get_streams()) != set():
            await asyncio.sleep(0.3)
            st = await session.get("http://localhost:5000/api/v1/config")
            state = EnsembleState.model_validate(await st.json())
            logging.debug(f"Waiting for lecroy ingester {state.get_streams()=}")

        map = {
            "lecroy": [
                [VirtualWorker(constraint=VirtualConstraint(i)).model_dump(mode="json")]
                for i in range(1, ntrig)
            ],
        }
        logging.debug(f"Sending {map=}")
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json=map,
        )
        assert resp.status == 200
        await resp.json()

    context = zmq.asyncio.Context()

    asyncio.create_task(
        stream_pkls(
            context,
            22004,
            PosixPath(filename),
            0.001,
            zmq.PUB,
        )
    )

    async with aiohttp.ClientSession() as session:
        st = await session.get("http://localhost:5000/api/v1/progress")
        content = await st.json()
        logging.debug(f"Progress {content=}")
        while not content["finished"]:
            await asyncio.sleep(0.3)
            st = await session.get("http://localhost:5000/api/v1/progress")
            content = await st.json()
            logging.debug(f"Progress {content=}")

        assert content == {
            "last_assigned": ntrig + 1,
            "completed_events": ntrig + 1,
            "total_events": ntrig + 1,
            "finished": True,
        }

    def work() -> None:
        f = h5pyd.File("http://localhost:5001/", "r")
        logging.info(
            f"file {f.keys()}",
        )
        assert f["max_val"][0] == max0
        assert f["ts"][0] == ts0
        assert f["max_val"].shape == (nburst,), "Unexpected shape of max_val result"
        assert f["ts"].shape == (nburst,), "Unexpected shape of ts result"

    loop = asyncio.get_event_loop()
    await loop.run_in_executor(None, work)



import asyncio
import logging

from dranspose.helpers.utils import cancel_and_wait
from dranspose.protocol import (
    WorkerName,
)

import pytest

from dranspose.worker import Worker, WorkerSettings
from tests.utils import wait_for_controller


@pytest.mark.asyncio
async def test_context(
    controller: None,
) -> None:
    worker = Worker(
        settings=WorkerSettings(
            worker_name=WorkerName("w5555"),
            worker_class="examples.repub.worker:RepubWorker",
        ),
    )
    worker_task = asyncio.create_task(worker.run())

    await wait_for_controller(workers={WorkerName("w5555")})

    logging.info("custom context is %s", worker.custom_context)
    assert worker.custom_context["context"].closed is False
    await worker.close()
    await cancel_and_wait(worker_task)
    logging.info("custom context is %s", worker.custom_context)
    assert worker.custom_context["context"].closed is True



import os
import pathlib
from pathlib import PosixPath

import asyncio
from typing import Awaitable, Callable, Any, Coroutine, Optional
import zmq.asyncio

import aiohttp
import pytest
from pydantic_core import Url

from dranspose.ingester import Ingester
from dranspose.ingesters.zmqsub_xspress3 import (
    ZmqSubXspressIngester,
    ZmqSubXspressSettings,
)
from dranspose.protocol import (
    StreamName,
    WorkerName,
    VirtualWorker,
    VirtualConstraint,
)
from dranspose.worker import Worker, WorkerSettings
from tests.utils import wait_for_controller, wait_for_finish


@pytest.mark.asyncio
async def test_multipart(
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[Worker], Awaitable[Worker]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    stream_pkls: Callable[
        [zmq.Context[Any], int, os.PathLike[Any] | str, float, int],
        Coroutine[Any, Any, None],
    ],
    tmp_path: pathlib.PurePath,
) -> None:
    await reducer("examples.dummy.reducer:FluorescenceReducer")
    await create_worker(
        Worker(
            settings=WorkerSettings(
                worker_name=WorkerName("w1"),
                worker_class="examples.dummy.worker:FluorescenceWorker",
            ),
        )
    )

    await create_ingester(
        ZmqSubXspressIngester(
            settings=ZmqSubXspressSettings(
                ingester_streams=[StreamName("xspress3")],
                upstream_url=Url("tcp://localhost:9999"),
                ingester_url=Url("tcp://localhost:10001"),
            ),
        )
    )

    await wait_for_controller(streams={StreamName("xspress3")})
    async with aiohttp.ClientSession() as session:
        ntrig = 5
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json={
                "xspress3": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(i)).model_dump(
                            mode="json"
                        )
                    ]
                    for i in range(ntrig)
                ],
            },
        )
        assert resp.status == 200
        await resp.json()

    context = zmq.asyncio.Context()

    asyncio.create_task(
        stream_pkls(
            context,
            9999,
            PosixPath("tests/data/xspress3-multipart.pkls"),
            0.001,
            zmq.PUB,
        )
    )

    content = await wait_for_finish()

    context.destroy()

    print(content)



import asyncio
import logging
import pickle
from collections import deque
from contextlib import asynccontextmanager
from typing import AsyncGenerator, Any

import zmq
from fastapi import FastAPI
from starlette.responses import Response

from dranspose.event import EventData
from dranspose.helpers.utils import done_callback, cancel_and_wait
from dranspose.protocol import WorkerUpdate, DistributedStateEnum, RedisKeys
from dranspose.worker import Worker

logger = logging.getLogger(__name__)


class DebugWorker(Worker):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.buffer: deque[EventData] = deque(maxlen=50)

    async def work(self) -> None:
        self._logger.info("started work task")
        await self.notify_worker_ready()

        while True:
            self.dequeue_task = None
            self.dequeue_task = asyncio.create_task(self.assignment_queue.get())
            ingesterset = await self.dequeue_task
            done = await self.poll_internals(ingesterset)
            if set(done) != {zmq.POLLIN}:
                self._logger.warning("not all sockets are pollIN %s", done)
                continue

            event = await self.build_event(ingesterset)
            self._logger.debug("adding event %s to buffer", event)
            self.buffer.append(event)

            wu = WorkerUpdate(
                state=DistributedStateEnum.IDLE,
                completed=[event.event_number],
                has_result=[False],
                worker=self.state.name,
            )
            await self.redis.xadd(
                RedisKeys.ready(self.state.mapping_uuid),
                {"data": wu.model_dump_json()},
            )
            self.state.processed_events += 1


worker: DebugWorker


@asynccontextmanager
async def lifespan(app: FastAPI) -> AsyncGenerator[None, None]:
    # Load the ML model
    global worker
    worker = DebugWorker()
    run_task = asyncio.create_task(worker.run())
    run_task.add_done_callback(done_callback)
    yield
    await cancel_and_wait(run_task)
    await worker.close()
    # Clean up the ML models and release the resources


app = FastAPI(lifespan=lifespan)


@app.get("/api/v1/status")
async def get_status() -> bool:
    return True


@app.get("/api/v1/last_events")
async def get_result(number: int = 1) -> Response:
    global worker
    print("getting ", number, "events")
    byteevs = []
    for i in range(-1, -number - 1, -1):
        try:
            lev = worker.buffer[i]
            byteevs.append(
                EventData(
                    event_number=lev.event_number,
                    streams={k: v.get_bytes() for k, v in lev.streams.items()},
                )
            )
        except Exception:
            pass
    data = pickle.dumps(list(reversed(byteevs)))
    return Response(data, media_type="application/x.pickle")



import asyncio
import hashlib
import importlib
import logging
import os
import sys
import traceback
from asyncio import Future, Task
from typing import Any

from dranspose.protocol import HashDigest, WorkParameter, ParameterName


def import_class(path: str) -> type:
    sys.path.append(os.getcwd())
    module = importlib.import_module(path.split(":")[0])
    custom = getattr(module, path.split(":")[1])
    return custom


def parameters_hash(parameters: dict[ParameterName, WorkParameter]) -> HashDigest:
    m = hashlib.sha256()
    for n in sorted(parameters):
        m.update(parameters[n].uuid.bytes)
    return HashDigest(m.hexdigest())


def done_callback(futr: Future[None]) -> None:
    try:
        futr.result()
    except asyncio.exceptions.CancelledError:
        pass
    except Exception as e:
        logging.error(
            "subroutine crashed %s trace: %s",
            e.__repr__(),
            traceback.format_exc(),
        )


async def cancel_and_wait(task: Task[Any] | Future[Any]) -> None:
    task.cancel()
    try:
        await task
    except asyncio.CancelledError:
        pass
    except StopIteration:
        pass
    except Exception as e:
        logging.error("cancel and wait task raised %s", e.__repr__())



import contextlib
import gzip
import json
import logging
import os
import pickle
import random
import threading
import time
import traceback
from typing import Iterator, Any, Optional, IO

import cbor2
import uvicorn
from fastapi import FastAPI, HTTPException
from pydantic import TypeAdapter
from pydantic_core import Url
from starlette.requests import Request
from starlette.responses import Response

from dranspose.helpers import utils
from dranspose.event import (
    InternalWorkerMessage,
    EventData,
    ResultData,
    message_tag_hook,
)
from dranspose.helpers.h5dict import router
from dranspose.helpers.jsonpath_slice_ext import NumpyExtentedJsonPathParser
from dranspose.protocol import (
    WorkerName,
    HashDigest,
    WorkParameter,
    ParameterName,
    ReducerState,
    WorkerState,
)


def get_internals(filename: os.PathLike[Any] | str) -> Iterator[InternalWorkerMessage]:
    f: IO[bytes]
    if str(filename).endswith(".gz"):
        f = gzip.open(filename, "rb")  # type: ignore[assignment]
        filename = str(filename)[:-3]
    else:
        f = open(filename, "rb")

    while True:
        try:
            if str(filename).endswith(".pkls"):
                frames = pickle.load(f)
            else:
                frames = cbor2.load(f, tag_hook=message_tag_hook)
            assert isinstance(frames, InternalWorkerMessage)
            yield frames
        except EOFError:
            break
    f.close()


logger = logging.getLogger(__name__)

ParamList = TypeAdapter(list[WorkParameter])

reducer_app = FastAPI()

reducer: Any | None = None


def get_data() -> dict[str, Any]:
    global reducer
    if reducer is not None and hasattr(reducer, "publish"):
        return reducer.publish
    return {}


reducer_app.state.get_data = get_data

reducer_app.include_router(router)


@reducer_app.get("/api/v1/result/{path:path}")
async def get_path(path: str) -> Any:
    global reducer
    if reducer is None or not hasattr(reducer, "publish"):
        raise HTTPException(status_code=404, detail="no publishable data")
    try:
        if path == "":
            path = "$"
        jsonpath_expr = NumpyExtentedJsonPathParser(debug=False).parse(path)
        print("expr", jsonpath_expr.__repr__())
        ret = [match.value for match in jsonpath_expr.find(reducer.publish)]
        data = pickle.dumps(ret)
        return Response(data, media_type="application/x.pickle")
    except Exception as e:
        raise HTTPException(status_code=400, detail="malformed path %s" % e.__repr__())


@reducer_app.post("/api/v1/parameter/{name}")
async def post_param(request: Request, name: ParameterName) -> HashDigest:
    data = await request.body()
    logger.info("got %s: %s (len %d)", name, data[:100], len(data))
    param = WorkParameter(name=name, data=data)
    if name in reducer_app.state.param_description:
        param_desc = reducer_app.state.param_description[name]
        param.value = param_desc.from_bytes(data)
    reducer_app.state.parameters[name] = param
    return HashDigest("")


@reducer_app.get("/api/v1/parameter/{name}")
async def get_param(name: ParameterName) -> Response:
    if name not in reducer_app.state.parameters:
        raise HTTPException(status_code=404, detail="Parameter not found")

    data = reducer_app.state.parameters[name].data
    return Response(data, media_type="application/x.bytes")


class Server(uvicorn.Server):
    def install_signal_handlers(self) -> None:
        pass

    @contextlib.contextmanager
    def run_in_thread(self, port: Optional[int]) -> Iterator[None]:
        if port is None:
            yield
            return
        thread = threading.Thread(target=self.run)
        thread.start()
        try:
            while not self.started:
                time.sleep(1e-3)
            yield
        finally:
            self.should_exit = True
            thread.join()


def get_parameters(
    parameter_file: Optional[os.PathLike[Any] | str], workercls: type, reducercls: type
) -> dict[ParameterName, WorkParameter]:
    parameters = {}
    logger.info("loading parameters from file %s", parameter_file)
    if parameter_file is not None:
        try:
            with open(parameter_file) as f:
                parameters = {
                    ParameterName(p.name): p for p in ParamList.validate_json(f.read())
                }
        except UnicodeDecodeError:
            logger.info("parameter file is not a json, try pickle")
            try:
                with open(parameter_file, "rb") as fb:
                    plist = pickle.load(fb)
                    parameters = {
                        ParameterName(p.name): p
                        for p in ParamList.validate_python(plist)
                    }
            except Exception as e:
                logger.warning(
                    "parameter file is not a pickle: %s, try cbor", e.__repr__()
                )
                with open(parameter_file, "rb") as fb:
                    plist = cbor2.load(fb)
                    parameters = {
                        ParameterName(p.name): p
                        for p in ParamList.validate_python(plist)
                    }

    logger.info("params from file %s", parameters)

    param_description = {}
    if hasattr(workercls, "describe_parameters"):
        param_description.update({p.name: p for p in workercls.describe_parameters()})
    if hasattr(reducercls, "describe_parameters"):
        param_description.update({p.name: p for p in reducercls.describe_parameters()})
    logger.info("parameter descriptions %s", param_description)

    for p in parameters:
        if p in param_description:
            parameters[p].value = param_description[p].from_bytes(parameters[p].data)
    logger.info("parsed params are %s", parameters)

    for p in param_description:
        if p not in parameters:
            parameters[p] = WorkParameter(
                name=p,
                value=param_description[p].default,
                data=param_description[p].to_bytes(param_description[p].default),
            )

    reducer_app.state.param_description = param_description
    logger.info("final params are %s", parameters)
    return parameters


def timer(red: Any) -> None:
    while True:
        delay = 1
        if hasattr(red, "timer"):
            delay = red.timer()

        time.sleep(delay)


def _work_event(
    worker: Any,
    index: int,
    reducer: Any,
    event: EventData,
    parameters: dict[ParameterName, WorkParameter],
    tick: bool,
) -> None:
    data = worker.process_event(event, parameters, tick)
    if data is None:
        return
    rd = ResultData(
        event_number=event.event_number,
        worker=WorkerName(f"replay-worker-{index}"),
        payload=data,
        parameters_hash=HashDigest(
            "688787d8ff144c502c7f5cffaafe2cc588d86079f9de88304c26b0cb99ce91c6"
        ),
    )

    header = rd.model_dump_json(exclude={"payload"}).encode("utf8")
    body = pickle.dumps(rd.payload)
    prelim = json.loads(header)
    prelim["payload"] = pickle.loads(body)
    result = ResultData.model_validate(prelim)

    reducer.process_result(result, parameters)


def _finish(
    workers: list[Any], reducer: Any, parameters: dict[ParameterName, WorkParameter]
) -> None:
    for worker in workers:
        if hasattr(worker, "finish"):
            try:
                worker.finish(parameters=parameters)
            except Exception as e:
                logger.error(
                    "worker finished failed with %s\n%s",
                    e.__repr__(),
                    traceback.format_exc(),
                )

    if hasattr(reducer, "finish"):
        try:
            reducer.finish(parameters=parameters)
        except Exception as e:
            logger.error(
                "reducer finish failed with %s\n%s",
                e.__repr__(),
                traceback.format_exc(),
            )


def replay(
    wclass: str,
    rclass: str,
    zmq_files: Optional[list[os.PathLike[Any] | str]] = None,
    source: Optional[str] = None,
    parameter_file: Optional[os.PathLike[Any] | str] = None,
    port: Optional[int] = None,
    stop_event: threading.Event | None = None,
    nworkers: int = 2,
    broadcast_first: bool = True,
    done_event: threading.Event | None = None,
    start_event: threading.Event | None = None,
    latency: float | None = None,
) -> None:
    if source is not None:
        sourcecls = utils.import_class(source)
        inst = sourcecls()
        gens = inst.get_source_generators()
    elif zmq_files is not None:
        gens = [get_internals(f) for f in zmq_files]
    else:
        gens = []

    workercls = utils.import_class(wclass)
    logger.info("custom worker class %s", workercls)

    reducercls = utils.import_class(rclass)
    logger.info("custom reducer class %s", reducercls)

    reducer_app.state.parameters = get_parameters(parameter_file, workercls, reducercls)

    logger.info("use parameters %s", reducer_app.state.parameters)

    global reducer
    workers = []
    for wi in range(nworkers):
        wstate = WorkerState(name=WorkerName(f"replay-worker-{wi}"))
        wobj = workercls(
            parameters=reducer_app.state.parameters, context={}, state=wstate
        )
        workers.append(wobj)

    rstate = ReducerState(url=Url("tcp://localhost:10200"))
    reducer = reducercls(
        parameters=reducer_app.state.parameters, context={}, state=rstate
    )

    logger.info("created workers %s", workers)
    threading.Thread(target=timer, daemon=True, args=(reducer,)).start()

    config = uvicorn.Config(
        reducer_app, port=port or 5000, host="localhost", log_level="info"
    )
    server = Server(config)
    # server.run()

    first = True

    with server.run_in_thread(port):
        cache = [None for _ in gens]
        last_tick = 0.0
        if start_event is not None:
            start_event.wait()
        while True:
            try:
                internals = [
                    next(gen) if ch is None else ch for gen, ch in zip(gens, cache)
                ]
                if len(internals) == 0:
                    break
                lowestevn = min([ev.event_number for ev in internals])
                lowinternals = []
                cache = internals
                for idx, ie in enumerate(internals):
                    if ie.event_number == lowestevn:
                        lowinternals.append(ie)
                        cache[idx] = None
                event = EventData.from_internals(lowinternals)

                dst_worker_ids = [random.randint(0, len(workers) - 1)]
                if first and broadcast_first:
                    dst_worker_ids = list(range(len(workers)))
                    first = False

                for wi in dst_worker_ids:
                    logger.info("ev %d spread to wi %d", event.event_number, wi)
                    tick = False
                    if hasattr(workers[wi], "get_tick_interval"):
                        interval_s = workers[wi].get_tick_interval(
                            parameters=reducer_app.state.parameters
                        )
                        if last_tick + interval_s < time.time():
                            tick = True
                            last_tick = time.time()
                    _work_event(
                        workers[wi],
                        wi,
                        reducer,
                        event,
                        reducer_app.state.parameters,
                        tick,
                    )
                if latency is not None:
                    time.sleep(latency)
            except StopIteration:
                logger.debug("end of replay, calling finish")
                _finish(workers, reducer, reducer_app.state.parameters)
                break
        if done_event is not None:
            done_event.set()
        logger.info("check if webserver should stay alive %s", stop_event)
        if stop_event is None:
            stop_event = threading.Event()
            stop_event.set()
        else:
            print("press ctrl-C to stop")
        try:
            logger.info("waiting for event")
            stop_event.wait()
        except KeyboardInterrupt:
            pass



import logging
import os
import subprocess
import time

import pytest


def test_simple() -> None:
    p = subprocess.Popen(["dranspose"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    time.sleep(4)
    p.terminate()
    text = p.communicate()[0]
    assert "usage: dranspose" in text.decode()


@pytest.mark.parametrize(
    "cmd, output",
    [
        (["controller"], "Uvicorn running on"),
        (
            ["reducer", "-c", "examples.dummy.reducer:FluorescenceReduce"],
            "created reducer with state service_uuid=UUID",
        ),
        (
            ["worker", "-c", "examples.dummy.worker:FluorescenceWorker"],
            "created worker with state service_uuid=UUID",
        ),
        (
            [
                "ingester",
                "-c",
                "TcpPcapIngester",
                "-u",
                "tcp://127.0.0.1:8888",
                "-n",
                "bla",
            ],
            "ingester:all subtasks running",
        ),
        (
            [
                "http_ingester",
                "-c",
                "dranspose.ingesters.http_sardana:app",
                "-n",
                "sardana",
            ],
            "Uvicorn running on",
        ),
        (["combined"], "worker2:registered ready message"),
        (["debugworker"], "Uvicorn running on"),
    ],
)
def test_component(cmd: list[str], output: str) -> None:
    p = subprocess.Popen(
        ["dranspose", *cmd], stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    time.sleep(4)
    p.terminate()
    text = b"".join(p.communicate())
    assert output in text.decode()


def test_replay_no_redis() -> None:
    my_env = os.environ.copy()
    my_env["REDIS_URL"] = "redis://nonexist-host:6379/0"
    p = subprocess.Popen(
        ["dranspose", "replay"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        env=my_env,
    )
    time.sleep(4)
    p.terminate()
    text = (b"".join(p.communicate())).decode()
    logging.info("got out+err: %s", text)
    assert "usage: dranspose" in text



import abc
import logging
import os
import time
from typing import Optional, Any
from importlib.metadata import version

import redis.asyncio as redis
from pydantic import UUID4, AliasChoices, Field, RedisDsn
from pydantic_core import Url
from pydantic_settings import BaseSettings

from dranspose.helpers.utils import parameters_hash, cancel_and_wait
from dranspose.parameters import Parameter, ParameterType
from dranspose.protocol import (
    RedisKeys,
    ControllerUpdate,
    IngesterState,
    WorkerState,
    ReducerState,
    ParameterName,
    WorkParameter,
    BuildGitMeta,
    StreamName,
)
import redis.exceptions as rexceptions
import asyncio


class DistributedSettings(BaseSettings):
    """
    Basic settings for any distributed service

    Attributes:
        redis_dsn: URL of the common redis instance
    """

    redis_dsn: RedisDsn = Field(
        Url("redis://localhost:6379/0"),
        validation_alias=AliasChoices("service_redis_dsn", "redis_url"),
    )

    build_meta_file: Optional[os.PathLike[Any] | str] = "/etc/build_git_meta.json"


class DistributedService(abc.ABC):
    """
    Abstract class defining common functionality for all services.
    """

    def __init__(
        self,
        state: WorkerState | IngesterState | ReducerState,
        settings: Optional[DistributedSettings] = None,
    ):
        self._distributed_settings = settings
        if self._distributed_settings is None:
            self._distributed_settings = DistributedSettings()

        self.state: WorkerState | IngesterState | ReducerState = state

        if ":" in state.name:
            raise Exception("Worker name must not contain a :")
        # TODO: check for already existing query string
        self.redis = redis.from_url(
            f"{self._distributed_settings.redis_dsn}?decode_responses=True&protocol=3"
        )
        self.raw_redis = redis.from_url(
            f"{self._distributed_settings.redis_dsn}?protocol=3"
        )
        self._logger = logging.getLogger(f"{__name__}+{self.state.name}")
        self._logger.info("log handlers are %s", self._logger.handlers)
        try:
            if self._distributed_settings.build_meta_file is not None:
                with open(self._distributed_settings.build_meta_file) as fd:
                    build_data = BuildGitMeta.model_validate_json(fd.read())
                    self._logger.info("build meta is %s", build_data)
                    self.state.mapreduce_version = build_data
        except Exception as e:
            logging.info("cannot load build meta information: %s", e.__repr__())
        self.state.dranspose_version = version("dranspose")
        self._logger.info("running version %s", self.state.dranspose_version)
        self.parameters: dict[ParameterName, WorkParameter] = {}

    def get_category(self) -> str:
        if isinstance(self.state, IngesterState):
            category = "ingester"
        elif isinstance(self.state, WorkerState):
            category = "worker"
        elif isinstance(self.state, ReducerState):
            category = "reducer"
        else:
            raise NotImplementedError(
                "Distributed Service not implemented for your Service"
            )
        return category

    async def publish_config(self) -> None:
        self._logger.debug("publish config %s", self.state)
        async with self.redis.pipeline() as pipe:
            await pipe.setex(
                RedisKeys.config(self.get_category(), self.state.name),
                10,
                self.state.model_dump_json(),
            )
            if hasattr(self, "param_descriptions"):
                for p in self.param_descriptions:
                    self._logger.debug("register parameter %s", p)
                    try:
                        await pipe.setex(
                            RedisKeys.parameter_description(p.name),
                            10,
                            p.model_dump_json(),
                        )
                    except Exception as e:
                        self._logger.error(
                            "failed to register parameter %s", e.__repr__()
                        )

            await pipe.execute()

    async def register(self) -> None:
        """
        Background job in every distributed service to publish the service's configuration.
        It publishes the `state` every 1.3 seconds or faster if there are updates from the controller with a new trigger map or parameters.
        """
        latest = await self.redis.xrevrange(RedisKeys.updates(), count=1)
        last = 0
        if len(latest) > 0:
            last = latest[0][0]
        while True:
            await self.publish_config()
            try:
                update_msgs = await self.redis.xread(
                    {RedisKeys.updates(): last}, block=1300
                )
                if RedisKeys.updates() in update_msgs:
                    update_msg = update_msgs[RedisKeys.updates()][0][-1]
                    last = update_msg[0]
                    update = ControllerUpdate.model_validate_json(update_msg[1]["data"])
                    self._logger.debug("update type %s", update)
                    newuuid = update.mapping_uuid
                    if newuuid != self.state.mapping_uuid:
                        self._logger.info(
                            "resetting config to %s with streams %s",
                            newuuid,
                            update.active_streams,
                        )
                        try:
                            await self.restart_work(newuuid, update.active_streams)
                        except Exception as e:
                            self._logger.error("restart_work failed %s", e.__repr__())
                    paramuuids = update.parameters_version
                    for name in paramuuids:
                        if (
                            name not in self.parameters
                            or self.parameters[name] != paramuuids[name]
                        ):
                            try:
                                params = await self.raw_redis.get(
                                    RedisKeys.parameters(name, paramuuids[name])
                                )
                                self._logger.debug(
                                    "received binary parameters for %s", name
                                )
                                if params is not None:
                                    self._logger.info(
                                        "set parameter %s of length %s",
                                        name,
                                        len(params),
                                    )
                                    self.parameters[name] = WorkParameter(
                                        name=name, uuid=paramuuids[name], data=params
                                    )
                                    # check if this parameter has a description and type
                                    desc = await self.redis.get(
                                        RedisKeys.parameter_description(name),
                                    )
                                    self._logger.debug("description is %s", desc)
                                    if desc:
                                        param_desc: ParameterType = (
                                            Parameter.validate_json(desc)
                                        )
                                        self._logger.info(
                                            "set paremter has a description %s",
                                            param_desc,
                                        )
                                        self.parameters[
                                            name
                                        ].value = param_desc.from_bytes(params)
                                        self._logger.debug(
                                            "parsed parameter value %s",
                                            self.parameters[name].value,
                                        )
                            except Exception as e:
                                self._logger.error(
                                    "failed to get parameters %s", e.__repr__()
                                )
                    self.state.parameters_hash = parameters_hash(self.parameters)
                    self._logger.debug(
                        "set local parameters to %s with hash %s",
                        {n: p.uuid for n, p in self.parameters.items()},
                        self.state.parameters_hash,
                    )
                    if update.finished:
                        self._logger.info("finished messages")
                        await self.finish_work()

            except rexceptions.ConnectionError:
                break
            except asyncio.exceptions.CancelledError:
                break

    async def multiprocess_run(self, queue: Any) -> None:
        task = asyncio.create_task(self.run())
        loop = asyncio.get_event_loop()
        await loop.run_in_executor(None, queue.get)
        await self.close()
        await cancel_and_wait(task)

    def sync_run(self, queue: Any) -> None:
        asyncio.run(self.multiprocess_run(queue))

    async def update_metrics(self) -> None:
        while True:
            start = time.time()
            old = self.state.processed_events
            await asyncio.sleep(1.0)
            end = time.time()
            rate = (self.state.processed_events - old) / (end - start)
            self.state.event_rate = rate
            if (self.state.processed_events - old) > 0:
                self._logger.info("receiving %lf frames per second", rate)

    async def run(self) -> None:
        pass

    async def restart_work(self, uuid: UUID4, active_streams: list[StreamName]) -> None:
        pass

    async def finish_work(self) -> None:
        pass

    async def close(self) -> None:
        await self.redis.delete(RedisKeys.config(self.get_category(), self.state.name))
        self._logger.info(
            "deleted redis key %s",
            RedisKeys.config(self.get_category(), self.state.name),
        )
        await self.redis.aclose()
        await self.raw_redis.aclose()



import asyncio
import logging
import struct
import time
from typing import Callable, Any, Coroutine, Optional

import pytest
import zmq.asyncio


async def sink(port: int | str, n: int) -> None:
    c = zmq.asyncio.Context()
    s = c.socket(zmq.PULL)
    if type(port) is str:
        s.connect(port)
    else:
        s.connect(f"tcp://localhost:{port}")
    start = time.perf_counter()
    total_bytes = 0
    for i in range(n):
        # logging.info("recv %d", i)
        data = await s.recv_multipart(copy=True)
        total_bytes += sum(map(len, data))
        if i % 100 == 0:
            end = time.perf_counter()
            logging.info(
                "sink 100 packets took %s: %lf p/s", end - start, 100 / (end - start)
            )
            start = end
        # logging.info("i %d, data %s", i, data)
    logging.info("total bytes %d", total_bytes)
    c.destroy()


async def time_sink(port: int, n: int) -> None:
    c = zmq.asyncio.Context()
    s = c.socket(zmq.PULL)
    s.connect(f"tcp://localhost:{port}")
    latencies = []
    for i in range(n):
        # logging.info("recv %d", i)
        data = await s.recv_multipart(copy=True)
        now = time.perf_counter()
        start = struct.unpack(">d", data[0])[0]
        # logging.info("latency was %lf", now-start)
        latencies.append(now - start)
        if i % 1000 == 0 and i != 0:
            logging.info(
                "last 1000 latency: avg %lf, max: %lf min: %lf",
                sum(latencies) / len(latencies),
                max(latencies),
                min(latencies),
            )
            latencies = []
            # end = time.perf_counter()
            # logging.info("sink 100 packets took %s: %lf p/s", end - start, 100 / (end - start))
            # start = end
        # logging.info("i %d, data %s", i, data)
    c.destroy()


async def forward(inport: int, outport: int, n: int) -> None:
    c = zmq.asyncio.Context()
    sin = c.socket(zmq.PULL)
    sout = c.socket(zmq.PUSH)
    sin.connect(f"tcp://localhost:{inport}")
    sout.bind(f"tcp://*:{outport}")
    start = time.perf_counter()
    for i in range(n):
        data = await sin.recv_multipart(copy=False)
        await sout.send_multipart(data)
        if i % 1000 == 0:
            end = time.perf_counter()
            logging.info(
                "forward 1000 packets took %s: %lf p/s",
                end - start,
                1000 / (end - start),
            )
            start = end
    c.destroy()


@pytest.mark.asyncio
async def test_zmq_rate_tcp(
    stream_small: Callable[
        [zmq.Context[Any], int, int, Optional[float]], Coroutine[Any, Any, None]
    ]
) -> None:
    ctask = asyncio.create_task(sink(9999, 1002))

    context = zmq.asyncio.Context()
    await stream_small(context, 9999, 1000, 0.000001)
    await ctask

    context.destroy()


@pytest.mark.asyncio
async def test_zmq_rate_ipc(
    stream_small: Callable[
        [zmq.Context[Any], str, int, Optional[float]], Coroutine[Any, Any, None]
    ]
) -> None:
    ctask = asyncio.create_task(sink("ipc:///tmp/zmq-test-0", 1002))

    context = zmq.asyncio.Context()
    await stream_small(context, "ipc:///tmp/zmq-test-0", 1000, 0.000001)
    await ctask

    context.destroy()


@pytest.mark.asyncio
async def test_zmq_rate_forward(
    stream_small: Callable[
        [zmq.Context[Any], int, int, Optional[float]], Coroutine[Any, Any, None]
    ]
) -> None:
    asyncio.create_task(forward(9999, 9998, 1002))
    ctask = asyncio.create_task(sink(9998, 1002))

    context = zmq.asyncio.Context()
    await stream_small(context, 9999, 1000, 0.00001)
    await ctask

    context.destroy()


@pytest.mark.asyncio
async def test_zmq_latency(
    time_beacon: Callable[
        [zmq.Context[Any], int, int, Optional[float]], Coroutine[Any, Any, None]
    ]
) -> None:
    ctask = asyncio.create_task(time_sink(9999, 10000))

    context = zmq.asyncio.Context()
    await time_beacon(context, 9999, 10000, 0.00001)
    await ctask
    context.destroy()


@pytest.mark.asyncio
async def test_zmq_forward_latency(
    time_beacon: Callable[
        [zmq.Context[Any], int, int, Optional[float]], Coroutine[Any, Any, None]
    ]
) -> None:
    asyncio.create_task(forward(9999, 9998, 10000))
    ctask = asyncio.create_task(time_sink(9998, 10000))

    context = zmq.asyncio.Context()
    await time_beacon(context, 9999, 10000, 0.00001)
    await ctask
    context.destroy()


async def route(inport: int, outport: int, n: int, workers: list[bytes]) -> None:
    c = zmq.asyncio.Context()
    sin = c.socket(zmq.PULL)
    sout = c.socket(zmq.ROUTER)
    sin.connect(f"tcp://localhost:{inport}")
    sout.bind(f"tcp://*:{outport}")
    for _ in workers:
        regdata = await sout.recv_multipart()
        logging.info("registered worker %s", regdata)
    start = time.perf_counter()
    for i in range(n):
        data = await sin.recv_multipart(copy=False)
        dest = zmq.Frame(workers[i % len(workers)])
        await sout.send_multipart([dest] + data)
        if i % 1000 == 0:
            end = time.perf_counter()
            logging.info(
                "forward 1000 packets took %s: %lf p/s",
                end - start,
                1000 / (end - start),
            )
            start = end
    c.destroy()


async def dealer_sink(port: int, name: bytes, n: int) -> None:
    c = zmq.asyncio.Context()
    s = c.socket(zmq.DEALER)
    s.setsockopt(zmq.IDENTITY, name)
    s.connect(f"tcp://localhost:{port}")
    await s.send(b"ping")
    start = time.perf_counter()
    total_bytes = 0
    for i in range(n):
        # logging.info("recv %d", i)
        data = await s.recv_multipart(copy=True)
        total_bytes += sum(map(len, data))
        if i % 1000 == 0:
            end = time.perf_counter()
            logging.info(
                "%s: sink 100 packets took %s: %lf p/s",
                name,
                end - start,
                1000 / (end - start),
            )
            start = end
        # logging.info("i %d, data %s", i, data)
    logging.info("total bytes %d", total_bytes)
    c.destroy()


@pytest.mark.asyncio
async def test_zmq_full_latency(
    time_beacon: Callable[
        [zmq.Context[Any], int, int, Optional[float]], Coroutine[Any, Any, None]
    ]
) -> None:
    asyncio.create_task(route(9999, 9998, 10000, [b"w1", b"w2"]))
    c1task = asyncio.create_task(dealer_sink(9998, b"w1", 5000))
    c2task = asyncio.create_task(dealer_sink(9998, b"w2", 5000))

    context = zmq.asyncio.Context()
    await time_beacon(context, 9999, 10000, 0.000001)
    await c1task
    await c2task
    context.destroy()



from typing import List
import numpy as np
import pickle
import logging

import zmq

from dranspose.data.lecroy import (
    LECROY_TYPE,
    LecroyBase,
    LecroyData,
    LecroyEnd,
    LecroyPacket,
    LecroyParsed,
    LecroyPrepare,
    LecroySeqStart,
)
from dranspose.event import StreamData

logger = logging.getLogger(__name__)


def parse(data: StreamData) -> LecroyBase:
    """
    Parses a lecroy packet

    Arguments:
        data: one or more frames with lecroy data

    Returns:
        either the original control packet or a LecroyParsed packet
    """

    assert data.typ == LECROY_TYPE
    assert data.length >= 1, "wrong number of multiparts"
    start = data.frames[0]
    if isinstance(start, zmq.Frame):
        start = start.bytes
    start_pkt = LecroyPacket.validate_json(start)
    logger.debug(f"lecroy packet {start_pkt=}")

    if isinstance(start_pkt, LecroyPrepare) or isinstance(start_pkt, LecroyEnd):
        return start_pkt

    assert isinstance(
        start_pkt, LecroySeqStart
    ), f"Cannot identify header frame: {start_pkt}"
    nch = len(start_pkt.channels)
    assert (
        data.length == 1 + nch * 3 + 1
    ), f"Unexpected number of frames ({data.length}) for {nch} channels"

    res = LecroyParsed(**start_pkt.model_dump())
    for ich in range(nch):
        offset = 1 + ich * 3
        metaframe = data.frames[offset]
        if isinstance(metaframe, zmq.Frame):
            metaframe = metaframe.bytes
        meta_pkt = LecroyPacket.validate_json(metaframe)
        logger.debug(f"lecroy packet {meta_pkt=}")
        assert isinstance(
            meta_pkt, LecroyData
        ), f"Cannot identify trace metadata frame: {meta_pkt}"
        res.meta.append(meta_pkt)

        tracesframe = data.frames[offset + 1]
        if isinstance(tracesframe, zmq.Frame):
            tracesframe = tracesframe.bytes
        buf = np.frombuffer(tracesframe, dtype=meta_pkt.dtype)
        traces = buf.reshape(meta_pkt.shape)
        res.data.append(traces)

        tsframe = data.frames[offset + 2]
        if isinstance(tsframe, zmq.Frame):
            tsframe = tsframe.bytes
        ts: List[float] = pickle.loads(tsframe)
        res.timestamps.append(ts)

    return res



import asyncio
import logging
from typing import Callable, Any, Coroutine, Optional, Awaitable

import aiohttp
import pytest
import zmq.asyncio

from dranspose.protocol import (
    VirtualWorker,
    VirtualConstraint,
    WorkerName,
    StreamName,
)
from dranspose.worker import Worker
from tests.utils import wait_for_controller, wait_for_finish

rust = pytest.mark.skipif("not config.getoption('rust')")


@pytest.mark.skip
@pytest.mark.asyncio
async def test_rust_basic(
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[WorkerName], Awaitable[Worker]],
    stream_small: Callable[
        [zmq.Context[Any], int, int, Optional[float]], Coroutine[Any, Any, None]
    ],
) -> None:
    await reducer(None)
    await create_worker(WorkerName("w1"), subprocess=True)  # type: ignore[call-arg]
    await create_worker(WorkerName("w2"), subprocess=True)  # type: ignore[call-arg]
    await create_worker(WorkerName("w3"), subprocess=True)  # type: ignore[call-arg]
    await create_worker(WorkerName("w4"), subprocess=True)  # type: ignore[call-arg]
    await create_worker(WorkerName("w5"), subprocess=True)  # type: ignore[call-arg]
    await create_worker(WorkerName("w6"), subprocess=True)  # type: ignore[call-arg]
    await create_worker(WorkerName("w7"), subprocess=True)  # type: ignore[call-arg]
    await create_worker(WorkerName("w8"), subprocess=True)  # type: ignore[call-arg]
    await create_worker(WorkerName("w9"), subprocess=True)  # type: ignore[call-arg]
    await create_worker(WorkerName("w10"), subprocess=True)  # type: ignore[call-arg]

    await asyncio.sleep(2)

    await wait_for_controller(streams={StreamName("eiger")})
    async with aiohttp.ClientSession() as session:
        logging.info("startup done")

        ntrig = 50000
        mapping = {
            "eiger": [
                [
                    VirtualWorker(constraint=VirtualConstraint(i // 10)).model_dump(
                        mode="json"
                    )
                ]
                for i in range(1, ntrig)
            ],
        }
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json=mapping,
        )
        assert resp.status == 200

    context = zmq.asyncio.Context()
    await stream_small(context, 9999, ntrig - 1, 0.000001)

    await wait_for_finish()

    context.destroy()



import pickle
from typing import Literal

import zmq
from pydantic import BaseModel, TypeAdapter, ConfigDict

from dranspose.event import StreamData


class ContrastBase(BaseModel):
    def to_stream_data(self) -> StreamData:
        dat = pickle.dumps(self.model_dump())
        return StreamData(typ="contrast", frames=[zmq.Frame(dat)])


class ContrastStarted(ContrastBase):
    """
    Example:
        ``` py
        ContrastStarted(
            status='started',
            path='/data/.../diff_1130_stream_test/raw/dummy',
            scannr=2,
            description='mesh sx -2 2 3 sy -2 2 4 0.1',
            snapshot={'attenuator1_x': 0.048,
                      'attenuator2_x': 0.067,
                      'attenuator3_x': 0.536,
                      'vfm_yaw': 0.15148492851039919,
                      'xrf_x': 94.99875}
        )
        ```
    """

    model_config = ConfigDict(extra="allow")

    status: Literal["started"] = "started"
    path: str
    scannr: int
    description: str


class ContrastRunning(ContrastBase):
    """
    Example:
        ``` py
        ContrastRunning(
            status='running',
            dt=2.410903215408325,
            sx=-2.000431059888797,
            sy=-2.0011940002441406,
            pseudo={
                'x': array([-2.00405186]),
                'y': array([-2.00290304]),
                'z': array([0.00029938]),
                'analog_x': array([-1.99962707]),
                'analog_y': array([-1.99349905]),
                'analog_z': array([-0.00306218])
            },
            panda0={
                'COUNTER1.OUT_Value': array([0.]),
                'COUNTER2.OUT_Value': array([0.]),
                'COUNTER3.OUT_Value': array([0.]),
                'FMC_IN.VAL6_Mean': array([0.04644599]),
                'FMC_IN.VAL7_Mean': array([-0.02943451]),
                'FMC_IN.VAL8_Mean': array([0.01255371])
            },
            xspress3={
                'type': 'Link',
                'filename': '/data/.../diff_1130_stream_test/raw/dummy/scan_000002_xspress3.hdf5',
                'path': '/entry/instrument/xspress3/',
                'universal': True
            }
        )
        ```
    """

    model_config = ConfigDict(extra="allow")

    status: Literal["running"] = "running"
    dt: float


class ContrastFinished(ContrastBase):
    """
    Example:
        ```python
        ContrastFinished(
            status='finished',
            path='/data/.../diff_1130_stream_test/raw/dummy',
            scannr=2,
            description='mesh sx -2 2 3 sy -2 2 4 0.1',
            snapshot={'attenuator1_x': 0.048,
                      'attenuator2_x': 0.066,
                      'vfm_yaw': 0.15148492851039919,
                      'xrf_x': 94.99875}
        )
        ```
    """

    model_config = ConfigDict(extra="allow")

    status: Literal["finished"] = "finished"
    path: str
    scannr: int
    description: str


class ContrastHeartbeat(ContrastBase):
    """
    Heartbeat message
    """

    status: Literal["heartbeat"]


ContrastPacket = TypeAdapter(
    ContrastStarted | ContrastRunning | ContrastFinished | ContrastHeartbeat
)
"""
A union type for contrast packets
"""



import logging

import pytest


import redis.asyncio as redis


@pytest.mark.asyncio
async def test_distributed() -> None:
    r = redis.Redis(host="localhost", port=6379, decode_responses=True, protocol=3)
    raw_redis = redis.from_url("redis://localhost:6379/0?protocol=3")

    keys = await r.keys("dranspose:*")
    logging.info("keys %s", keys)

    key = await r.set("dranspose:testkey", b"")
    logging.info("key set to %s", key)
    assert key is True

    params = await raw_redis.get("dranspose:testkey")
    logging.info("param is %s", params)
    assert params is not None
    keys = await r.keys("dranspose:*")
    logging.info("keys %s", keys)

    await r.delete("dranspose:testkey")

    await r.aclose()
    await raw_redis.aclose()



import asyncio
import logging
from typing import Awaitable, Callable, Any, Coroutine, Optional

import aiohttp

import pytest
import zmq.asyncio
import zmq
from pydantic_core import Url

from dranspose.ingester import Ingester
from dranspose.ingesters.zmqpull_single import (
    ZmqPullSingleIngester,
    ZmqPullSingleSettings,
)
from dranspose.protocol import (
    StreamName,
    WorkerName,
    VirtualWorker,
    VirtualConstraint,
    EnsembleState,
)

from dranspose.worker import Worker
from tests.utils import wait_for_controller


@pytest.mark.asyncio
async def test_simple(
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[WorkerName], Awaitable[Worker]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    stream_small: Callable[
        [zmq.Context[Any], int, int, Optional[float]], Coroutine[Any, Any, None]
    ],
) -> None:
    await reducer(None)
    await create_worker(WorkerName("w1"), subprocess=True)  # type: ignore[call-arg]
    await create_worker(WorkerName("w2"), subprocess=True)  # type: ignore[call-arg]
    await create_worker(WorkerName("w3"), subprocess=True)  # type: ignore[call-arg]
    await create_worker(WorkerName("w4"), subprocess=True)  # type: ignore[call-arg]
    await create_worker(WorkerName("w5"), subprocess=True)  # type: ignore[call-arg]
    await create_worker(WorkerName("w6"), subprocess=True)  # type: ignore[call-arg]
    await create_worker(WorkerName("w7"), subprocess=True)  # type: ignore[call-arg]
    await create_worker(WorkerName("w8"), subprocess=True)  # type: ignore[call-arg]
    await create_worker(WorkerName("w9"), subprocess=True)  # type: ignore[call-arg]
    await create_worker(WorkerName("w10"), subprocess=True)  # type: ignore[call-arg]
    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("eiger")],
                upstream_url=Url("tcp://localhost:9999"),
            ),
        ),
        subprocess=True,  # type: ignore[call-arg]
    )

    await wait_for_controller(streams={StreamName("eiger")})
    async with aiohttp.ClientSession() as session:
        ntrig = 100
        mapping = {
            "eiger": [
                [
                    VirtualWorker(constraint=VirtualConstraint(i // 10)).model_dump(
                        mode="json"
                    )
                ]
                for i in range(1, ntrig)
            ],
        }
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json=mapping,
        )
        assert resp.status == 200

    context = zmq.asyncio.Context()

    asyncio.create_task(stream_small(context, 9999, ntrig - 1, 0.000001))

    async with aiohttp.ClientSession() as session:
        st = await session.get("http://localhost:5000/api/v1/progress")
        content = await st.json()
        while not content["finished"]:
            await asyncio.sleep(0.9)
            st = await session.get("http://localhost:5000/api/v1/progress")
            content = await st.json()
            st = await session.get("http://localhost:5000/api/v1/config")
            conf = EnsembleState.model_validate(await st.json())
            msg = []
            for k in conf.workers + conf.ingesters + [conf.reducer]:
                if k is not None:
                    msg.append(f"{k.name}:{k.processed_events} -- {k.event_rate}")
            logging.info("config is \n%s", "\n".join(msg))

        assert content == {
            "last_assigned": ntrig + 1,
            "completed_events": ntrig + 1,
            "total_events": ntrig + 1,
            "finished": True,
        }
    context.destroy()



from datetime import datetime
from typing import Optional

import zmq

from dranspose.data.positioncap import (
    PositionCapStart,
    PositionCapEnd,
    PositionCapValues,
    PositionCapField,
    PositionCapPacketType,
)
from dranspose.event import StreamData


class PositioncapParser:
    def __init__(self) -> None:
        self.arm_time: Optional[datetime] = None
        self.fields: list[PositionCapField] = []

    def parse(self, data: StreamData) -> PositionCapPacketType:
        """
        Parses a position capture packet

        Arguments:
            data: a frame comming from pcap

        Returns:
            a position capture
        """
        assert data.typ == "PCAP_RAW"
        assert data.length == 1
        frame = data.frames[0]
        if isinstance(frame, zmq.Frame):
            val = frame.bytes.decode()
        else:
            val = frame.decode()

        if val.startswith("arm_time:"):
            # header
            for line in val.split("\n"):
                if line.startswith("arm_time: "):
                    # TODO: removing the Z at the end of the isoformat is required for python 3.10, remove once support is dropped
                    self.arm_time = datetime.fromisoformat(
                        line[len("arm_time: ") :].replace("Z", "")
                    )
                if line.startswith(" "):
                    parts = line.strip().split(" ")
                    self.fields.append(
                        PositionCapField(name=f"{parts[0]}.{parts[2]}", type=parts[1])
                    )
            if self.arm_time is None:
                raise Exception("unable to parse header")
            return PositionCapStart(arm_time=self.arm_time)
        elif val.startswith("END"):
            return PositionCapEnd()
        else:
            if self.arm_time is None:
                raise Exception("unable to parse data, never got arm message")
            parts = val.strip().split(" ")
            ret = PositionCapValues()
            for f, v in zip(self.fields, parts):
                ret.fields[f.name] = f.parse(v)
            return ret



import logging

from dranspose.data.pcap import PCAPData
from dranspose.event import EventData
from dranspose.middlewares.pcap import parse

logger = logging.getLogger(__name__)


class PcapWorker:
    def __init__(self, *args, **kwargs):
        pass

    def process_event(self, event: EventData, *args, **kwargs):
        if "pcap" in event.streams:
            res = parse(event.streams["pcap"])
            if isinstance(res, PCAPData):
                logger.debug("parsed packet is %s", res)
            else:
                logger.info("control packet is %s", res)



import asyncio
import logging
import time
from typing import Callable, Any, Optional, Coroutine

import pytest
import zmq.asyncio


@pytest.mark.asyncio
async def test_zmq_multiroute(
    time_beacon: Callable[
        [zmq.Context[Any], int, int, Optional[float]], Coroutine[Any, Any, None]
    ]
) -> None:
    async def dealer_msink(ports: list[int], name: bytes, n: int) -> None:
        c = zmq.asyncio.Context()
        s = c.socket(zmq.DEALER)
        s.setsockopt(zmq.IDENTITY, name)
        for port in ports:
            s.connect(f"tcp://localhost:{port}")
        for _ in ports:
            await s.send(b"ping")
        start = time.perf_counter()
        total_bytes = 0
        for i in range(n):
            logging.debug("sink recv %d", i)
            data = await s.recv_multipart(copy=True)
            total_bytes += sum(map(len, data))
            if i % 1000 == 0:
                end = time.perf_counter()
                logging.info(
                    "%s: sink 100 packets took %s: %lf p/s",
                    name,
                    end - start,
                    1000 / (end - start),
                )
                start = end
            # logging.info("i %d, data %s", i, data)
        logging.info("total bytes %d", total_bytes)
        c.destroy()

    async def route(inport: int, outport: int, n: int, workers: list[bytes]) -> None:
        c = zmq.asyncio.Context()
        sin = c.socket(zmq.PULL)
        sout = c.socket(zmq.ROUTER)
        sin.connect(f"tcp://localhost:{inport}")
        sout.bind(f"tcp://*:{outport}")
        for _ in workers:
            regdata = await sout.recv_multipart()
            logging.info("%d: registered worker %s", inport, regdata)
        start = time.perf_counter()
        for i in range(n):
            data = await sin.recv_multipart(copy=False)
            await sout.send_multipart([workers[i % len(workers)]] + data)
            if i % 1000 == 0:
                end = time.perf_counter()
                logging.info(
                    "%d: forward 1000 packets took %s: %lf p/s",
                    inport,
                    end - start,
                    1000 / (end - start),
                )
                start = end
        c.destroy()

    pkts = 10000
    ctask = asyncio.create_task(dealer_msink([9998, 9918], b"w1", 2 * pkts))
    r1task = asyncio.create_task(route(9999, 9998, pkts, [b"w1"]))
    r2task = asyncio.create_task(route(9919, 9918, pkts, [b"w1"]))

    context = zmq.asyncio.Context()
    s1task = asyncio.create_task(time_beacon(context, 9999, pkts, 0.00001))
    s2task = asyncio.create_task(time_beacon(context, 9919, pkts, 0.00001))

    await s1task
    await s2task

    await ctask
    await r1task
    await r2task

    context.destroy()



import datetime
import itertools

import numpy as np

from dranspose.data.contrast import ContrastStarted, ContrastRunning, ContrastFinished
from dranspose.data.positioncap import (
    PositionCapStart,
    PositionCapField,
    PositionCapValues,
    PositionCapEnd,
)
from dranspose.data.xspress3 import XspressStart, XspressImage, XspressEnd
from dranspose.event import InternalWorkerMessage

from dranspose.data.stream1 import Stream1Start, Stream1Data, Stream1End

# import h5py
# from bitshuffle import compress_lz4


class FluorescenceSource:
    def __init__(self):
        # self.fd = h5py.File("../nanomax/000008.h5")
        pass

    def get_source_generators(self):
        return [
            self.pilatus_source(),
            self.contrast_source(),
            self.xspress3_source(),
            self.pcap_source(),
        ]

    def pilatus_source(self):
        msg_number = itertools.count(0)

        stins_start = Stream1Start(filename="", msg_number=next(msg_number))
        start = InternalWorkerMessage(
            event_number=0,
            streams={"pilatus": stins_start.to_stream_data()},
        )
        yield start

        frameno = 0
        # for image in self.fd["/entry/measurement/pilatus/frames"]:
        for _ in range(10):
            image = np.ones((10, 10))
            # dat = compress_lz4(image)
            stins = Stream1Data(
                msg_number=next(msg_number),
                frame=frameno,
                shape=image.shape,
                compression="none",  # compression="bslz4"
                type=str(image.dtype),
                data=image,
            )
            img = InternalWorkerMessage(
                event_number=frameno + 1,
                streams={"pilatus": stins.to_stream_data()},
            )
            yield img
            frameno += 1

        stins_end = Stream1End(msg_number=next(msg_number))
        end = InternalWorkerMessage(
            event_number=frameno + 1,
            streams={"pilatus": stins_end.to_stream_data()},
        )
        yield end

    def contrast_source(self):
        start = ContrastStarted(path="", scannr=4, description="stepscan")
        yield InternalWorkerMessage(
            event_number=0,
            streams={"contrast": start.to_stream_data()},
        )

        data = {}
        data["x"] = np.ones((10,))  # self.fd["entry/measurement/pseudo/x"][:]
        data["y"] = np.ones((10,))  # self.fd["entry/measurement/pseudo/y"][:]
        data["z"] = np.ones((10,))  # self.fd["entry/measurement/pseudo/z"][:]

        for i in range(len(data["x"])):
            pseudo = {k: np.array([v[i]]) for k, v in data.items()}
            print(pseudo)
            pos = ContrastRunning(pseudo=pseudo, dt=0.4)
            yield InternalWorkerMessage(
                event_number=i + 1,
                streams={"contrast": pos.to_stream_data()},
            )

        end = ContrastFinished(path="", scannr=4, description="stepscan")
        yield InternalWorkerMessage(
            event_number=11,
            streams={"contrast": end.to_stream_data()},
        )

    def xspress3_source(self):
        start = XspressStart(filename="")
        yield InternalWorkerMessage(
            event_number=0,
            streams={"xspress3": start.to_stream_data()},
        )

        frameno = 0
        # for spec in self.fd["/entry/measurement/xspress3/data"]:
        for _ in range(10):
            spec = np.ones((4, 10))
            dat = XspressImage(
                frame=frameno,
                shape=spec.shape,
                exptime=0.099999875,
                type=str(spec.dtype),
                compression="none",
                data=spec,
                meta={"ocr": 1},
            )
            yield InternalWorkerMessage(
                event_number=frameno + 1,
                streams={"xspress3": dat.to_stream_data()},
            )
            frameno += 1

        end = XspressEnd()
        yield InternalWorkerMessage(
            event_number=frameno + 1,
            streams={"xspress3": end.to_stream_data()},
        )

    def pcap_source(self):
        fields = [
            PositionCapField(name="PCAP.BITS0.Value", type="uint32"),
            PositionCapField(name="INENC1.VAL.Mean", type="double"),
            PositionCapField(name="PCAP.TS_TRIG.Value", type="double"),
        ]
        start = PositionCapStart(arm_time=datetime.datetime.now(datetime.timezone.utc))
        yield InternalWorkerMessage(
            event_number=0,
            streams={"pcap": start.to_stream_data(fields)},
        )

        data = [
            np.ones((10,), dtype=np.uint32),
            np.ones(
                (10,), dtype=np.float64
            ),  # self.fd["/entry/measurement/panda0/INENC1.VAL_Mean"][:],
            np.ones(
                (10,), dtype=np.float64
            ),  # self.fd["/entry/measurement/panda0/PCAP.TS_TRIG_Value"][:]
        ]

        for i in range(len(data[0])):
            for f, d in zip(fields, data):
                f.value = d[i]

            print(fields)
            val = PositionCapValues(fields={f.name: f for f in fields})
            yield InternalWorkerMessage(
                event_number=i + 1,
                streams={"pcap": val.to_stream_data()},
            )

        end = PositionCapEnd()
        yield InternalWorkerMessage(
            event_number=11,
            streams={"pcap": end.to_stream_data()},
        )



import base64
import gzip
import logging
import time
from typing import Any, Literal

import numpy as np
from fastapi import APIRouter, FastAPI, HTTPException
from starlette.requests import Request
from starlette.responses import Response

from dranspose.helpers.h5types import (
    H5Root,
    H5Group,
    H5Attribute,
    H5Shape,
    H5Type,
    H5IntType,
    H5FloatType,
    H5StrType,
    H5ValuedAttribute,
    H5ScalarShape,
    H5SimpleShape,
    H5CompType,
    H5NamedType,
    H5Link,
    H5Dataset,
    H5UUID,
)

router = APIRouter()

# only activate for running it directly, otherwise it overwrites cli.py basic config
# logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger()


def _path_to_uuid(path: list[str], collection_type: Literal["g-", "d-"]) -> H5UUID:
    abspath = "/" + "/".join(path)
    # print("abspath", abspath)
    uuid = base64.b16encode(abspath.encode()).decode()
    # print("uuid from function is", str(id(obj)).encode() + b"-" + uuid)
    return H5UUID(collection_type + "h5dict-" + uuid)


def _get_obj_at_path(
    data: dict[str, Any], path: str | list[str]
) -> tuple[Any, list[str]]:
    obj = data
    clean_path = []
    if isinstance(path, str):
        path = path.split("/")
    for p in path:
        # print("path part is:", p)
        if p == "":
            continue
        # print("traverse", obj)
        obj = obj[p]
        clean_path.append(p)
    return obj, clean_path


def _uuid_to_obj(data: dict[str, Any], uuid: str) -> tuple[Any, list[str]]:
    logger.debug("parse %s", uuid)
    col_type, idstr, path = uuid.split("-")
    path = base64.b16decode(path).decode()
    logger.debug("raw path %s", path)
    # assert id(data) == int(idstr)
    return _get_obj_at_path(data, path)


def _canonical_to_h5(canonical: str) -> H5Type | None:
    if canonical.startswith(">"):
        order = "BE"
    else:
        order = "LE"
    bytelen = int(canonical[2:])
    htyp: H5Type | None = None
    if canonical[1] == "f":
        # floating
        htyp = H5FloatType(base=f"H5T_IEEE_F{8 * bytelen}{order}")
    elif canonical[1] in ["u", "i"]:
        signed = canonical[1].upper()
        htyp = H5IntType(base=f"H5T_STD_{signed}{8 * bytelen}{order}")
    else:
        logger.error("numpy type %s not available", canonical)

    return htyp


def _make_shape_type(obj: Any) -> tuple[H5Shape | None, H5Type | None]:
    h5shape: H5Shape | None = None
    h5type: H5Type | None = None
    if type(obj) is int:
        h5shape = H5ScalarShape()
        h5type = H5IntType(base="H5T_STD_I64LE")
    elif type(obj) is float:
        h5shape = H5ScalarShape()
        h5type = H5FloatType(base="H5T_IEEE_F64LE")

    elif type(obj) is str:
        h5shape = H5ScalarShape()
        h5type = H5StrType()

    elif isinstance(obj, list) and len(obj) > 0 and isinstance(obj[0], str):
        h5shape = H5SimpleShape(dims=[len(obj)])
        h5type = H5StrType()
    else:
        try:
            arr = np.array(obj)
            h5shape = H5SimpleShape(dims=arr.shape)

            if len(arr.dtype.descr) > 1:
                # compound datatype
                fields = []
                for field in arr.dtype.descr:
                    typ = _canonical_to_h5(field[1])
                    if typ is not None:
                        fields.append(H5NamedType(name=field[0], type=typ))
                h5type = H5CompType(fields=fields)
            else:
                canonical = arr.dtype.descr[0][1]
                h5type = _canonical_to_h5(canonical)
            logging.debug("convert np dtype %s  to %s", arr.dtype, h5type)

        except Exception as e:
            logger.error("esception in handling code %s", e.__repr__())

    return h5shape, h5type


def _dataset_from_obj(
    data: Any, path: list[str], obj: Any, uuid: H5UUID
) -> H5Dataset | None:
    shape, typ = _make_shape_type(obj)
    if shape is not None and typ is not None:
        ret = H5Dataset(
            id=uuid,
            attributeCount=len(_get_obj_attrs(data, path)),
            shape=shape,
            type=typ,
        )
        return ret
    return None


def _get_group_link(obj: Any, path: list[str]) -> H5Link:
    collection: Literal["datasets", "groups"]
    collection_type: Literal["g-", "d-"]
    if isinstance(obj, dict):
        collection = "groups"
        collection_type = "g-"
    else:
        collection = "datasets"
        collection_type = "d-"
    return H5Link(
        collection=collection,
        id=_path_to_uuid(path, collection_type=collection_type),
        title=path[-1],
    )


def _get_group_links(obj: Any, path: list[str]) -> list[H5Link]:
    links = []
    if isinstance(obj, dict):
        for key, val in obj.items():
            if not isinstance(key, str):
                logger.warning("unable to use non-string key: %s as group name", key)
                continue
            if key.endswith("_attrs"):
                continue
            link = _get_group_link(val, path + [key])
            links.append(link)
    return links


def _get_attr(
    aobj: dict[str, Any], name: str, include_values: bool = True
) -> H5Attribute | H5ValuedAttribute | None:
    # print("get attribute")
    if name in aobj:
        h5shape, h5type = _make_shape_type(aobj[name])
        # print("shape and type ret", h5shape, h5type)
        if h5shape is not None and h5type is not None:
            ret = H5Attribute(name=name, shape=h5shape, type=h5type)
            if include_values:
                val = aobj[name]
                if isinstance(val, np.ndarray):
                    val = val.tolist()
                ret = H5ValuedAttribute(**ret.model_dump(by_alias=True), value=val)
            return ret
    return None


def _make_attrs(
    aobj: dict[str, Any], include_values: bool = False
) -> list[H5Attribute] | list[H5ValuedAttribute]:
    # print("make attributes of ", aobj)
    attrs = []
    if isinstance(aobj, dict):
        for key, value in aobj.items():
            if not isinstance(key, str):
                logger.warning(
                    "unable to use non-string key: %s as attribute name", key
                )
                continue
            attr = _get_attr(aobj, key, include_values=include_values)
            if attr is not None:
                attrs.append(attr)
    return attrs


def _get_obj_attrs(
    data: dict[str, Any], path: list[str], include_values: bool = False
) -> list[H5Attribute] | list[H5ValuedAttribute]:
    if len(path) == 0:
        # get root attributes
        if "_attrs" in data:
            return _make_attrs(data["_attrs"], include_values=include_values)
    else:
        # print("normal attr fetch")
        parent, _ = _get_obj_at_path(data, path[:-1])
        # print("parent is", parent)
        if f"{path[-1]}_attrs" in parent:
            # print( "make attrs for")
            return _make_attrs(
                parent[f"{path[-1]}_attrs"], include_values=include_values
            )
    return []


@router.get("/datasets/{uuid}/value", response_model=None)
def values(
    req: Request, uuid: H5UUID, select: str | None = None
) -> dict[str, Any] | Response:
    data = req.app.state.get_data()
    obj, _ = _uuid_to_obj(data, uuid)
    logger.debug("return value for obj %s", obj)
    logger.debug("selection %s", select)

    if type(obj) in [int, float, str]:
        return {"value": obj}

    elif isinstance(obj, list) and len(obj) > 0 and isinstance(obj[0], str):
        return {"value": obj}
    else:
        ret = np.array(obj)
        slices = []
        if select is not None:
            dims = select[1:-1].split(",")
            slices = [slice(*map(int, dim.split(":"))) for dim in dims]

        logger.debug("slices %s", slices)

        if len(slices) == 1:
            ret = ret[slices[0]]
        elif len(slices) > 1:
            # TODO: from python 3.11, this can be written as
            # ret = ret[*slices]
            ret = ret[tuple(slices)]

        ret_bytes = ret.tobytes()
        if "gzip" in req.headers.get("Accept-Encoding", ""):
            if len(ret_bytes) > 1000:
                compressed_data = gzip.compress(ret_bytes, compresslevel=7)

                return Response(
                    content=compressed_data,
                    media_type="application/octet-stream",
                    headers={"Content-Encoding": "gzip"},
                )
        return Response(content=ret_bytes, media_type="application/octet-stream")


@router.get("/datasets/{uuid}")
def datasets(req: Request, uuid: H5UUID) -> H5Dataset:
    data = req.app.state.get_data()
    obj, path = _uuid_to_obj(data, uuid)
    logger.debug("return dataset for obj %s", obj)
    ret = _dataset_from_obj(data, path, obj, uuid)
    # print("return dset", ret)
    if ret is None:
        raise HTTPException(status_code=404, detail="Dataset not found")
    return ret


@router.get("/groups/{uuid}/links/{name}")
def link(req: Request, uuid: H5UUID, name: str) -> dict[Literal["link"], H5Link]:
    data = req.app.state.get_data()
    obj, path = _uuid_to_obj(data, uuid)
    path.append(name)
    if name in obj:
        obj = obj[name]
        key: Literal["link"] = "link"
        ret = {key: _get_group_link(obj, path)}
        logger.debug("link name is %s", ret)
        return ret
    raise HTTPException(status_code=404, detail="Link not found")


@router.get("/groups/{uuid}/links")
def links(req: Request, uuid: H5UUID) -> dict[Literal["links"], list[H5Link]]:
    data = req.app.state.get_data()
    obj, path = _uuid_to_obj(data, uuid)
    key: Literal["links"] = "links"
    ret = {key: _get_group_links(obj, path)}
    logger.debug("group links %s", ret)
    return ret


@router.get("/{typ}/{uuid}/attributes/{name}")
def attribute(
    req: Request, typ: Literal["groups", "datasets"], uuid: H5UUID, name: str
) -> H5ValuedAttribute:
    # print("get attr with name", typ, uuid, name)
    data = req.app.state.get_data()
    obj, path = _uuid_to_obj(data, uuid)
    logger.debug(
        "start listing attributes typ %s id %s, obj %s, path: %s", typ, uuid, obj, path
    )
    allattrs = _get_obj_attrs(data, path, include_values=True)
    for attr in allattrs:
        if attr.name == name:
            logger.debug("return attribute %s", attr)
            if isinstance(attr, H5ValuedAttribute):
                return attr
    raise HTTPException(status_code=404, detail="Attribute not found")


@router.get("/{typ}/{uuid}/attributes")
def attributes(
    req: Request, typ: Literal["groups", "datasets"], uuid: H5UUID
) -> dict[Literal["attributes"], list[H5Attribute]]:
    data = req.app.state.get_data()
    obj, path = _uuid_to_obj(data, uuid)
    logger.debug(
        "start listing attributes typ %s id %s, obj %s, path: %s", typ, uuid, obj, path
    )
    # by calling _get_obj_attrs with False, it should never return H5ValuedAttribute
    attrs: list[H5Attribute] = _get_obj_attrs(data, path, include_values=False)  # type: ignore[assignment]
    return {"attributes": attrs}


@router.get("/groups/{uuid}")
def group(req: Request, uuid: H5UUID) -> H5Group:
    data = req.app.state.get_data()
    obj, path = _uuid_to_obj(data, uuid)
    logger.debug("start listing group id %s, obj %s, path: %s", uuid, obj, path)

    if isinstance(obj, dict):
        linkCount = len(
            list(
                filter(
                    lambda x: isinstance(x, str) and not x.endswith("_attrs"),
                    obj.keys(),
                )
            )
        )
        group = H5Group(
            root=_path_to_uuid([], collection_type="g-"),
            id=uuid,
            linkCount=linkCount,
            attributeCount=len(_get_obj_attrs(data, path)),
        )
        logger.debug("group is %s", group)
        return group
    raise HTTPException(status_code=404, detail="Group not found")


@router.get("/")
def read_root(request: Request) -> H5Root:
    logging.debug("data %s", request.app.state.get_data())
    data = request.app.state.get_data()
    if isinstance(data, dict):
        uuid = _path_to_uuid([], collection_type="g-")
        ret = H5Root(root=uuid)
        return ret
    raise Exception("data object is not a dict")


app = FastAPI()

app.include_router(router)


def get_data() -> dict[str, Any]:
    dt = np.dtype({"names": ["a", "b"], "formats": [float, int]})

    arr = np.array([(0.5, 1)], dtype=dt)

    changing = np.ones((5, int(time.time()) % 10 + 5))
    return {
        "map": {
            "x": [
                np.float64(-14.96431272),
                np.float64(-14.76401095),
                np.float64(-14.56276384),
                np.float64(-14.361908),
                np.float64(-14.16112703),
            ],
            "y": [
                np.float64(-10.00637736),
                np.float64(-10.00502708),
                np.float64(-10.00403313),
                np.float64(-10.00349819),
                np.float64(-10.00320074),
            ],
            "values": [
                np.float32(0.6831444),
                np.float32(0.0),
                np.float32(0.039953336),
                np.float32(0.14946304),
                np.float32(0.0),
            ],
        },
        "live": 34,
        "other": {"third": [1, 2, 3]},  # only _attrs allowed in root
        "other_attrs": {"NX_class": "NXother"},
        "image": np.ones((1000, 1000)),
        "image_attrs": {"listattr": [42, 43, 44, 45]},
        "changing_shape": changing,
        "specialtyp": np.ones((10, 10), dtype=">u8"),
        "specialtyp_attrs": {"NXdata": "NXspecial"},
        "hello": "World",
        "_attrs": {"NX_class": "NXentry"},
        "composite": arr,
        "composite_attrs": {"axes": ["I", "q"]},
        "oned_attrs": {"NX_class": "NXdata", "axes": ["motor"], "signal": "data"},
        "oned": {
            "data": np.ones((42)),
            "data_attrs": {"long_name": "photons"},
            "motor": np.linspace(0, 1, 42),
            "motor_attrs": {"long_name": "motor name"},
        },
        "twod_attrs": {
            "NX_class": "NXdata",
            "axes": ["motor", "."],
            "signal": "frame",
            "interpretation": "image",
        },
        "twod": {
            "frame": np.ones((42, 42)),
            "motor": np.linspace(0, 1, 42),
            "motor_attrs": {"long_name": "motor name"},
        },
    }


app.state.get_data = get_data



import logging

import zmq

from dranspose.event import EventData

logger = logging.getLogger(__name__)


class RepubWorker:
    def __init__(self, context, state, **kwargs):
        logger.info("context is %s", context)
        logger.info("state is %s", state)
        if "context" not in context:
            context["context"] = zmq.Context()
            context["socket"] = context["context"].socket(zmq.PUSH)
            # hack to extract port from worker name
            port = int(state.name[1:])
            context["socket"].bind(f"tcp://0.0.0.0:{port}")

        self.sock = context["socket"]
        self.buffer = {}

    def process_event(self, event: EventData, parameters=None, *args, **kwargs):
        logger.debug("using parameters %s", parameters)
        logger.debug("event %s", event.streams["eiger"])

        self.buffer[event.event_number] = event

        self.sock.send_multipart(event.streams["eiger"].frames)

    def close(self, context):
        if "socket" in context:
            context["socket"].close()
        if "context" in context:
            context["context"].destroy(linger=0)



from typing import Awaitable, Callable, Any, Coroutine, Optional

import aiohttp

import pytest
import zmq
from pydantic_core import Url

from dranspose.ingester import Ingester
from dranspose.ingesters.zmqpull_single import (
    ZmqPullSingleIngester,
    ZmqPullSingleSettings,
)
from dranspose.protocol import (
    StreamName,
    WorkerName,
)

from dranspose.worker import Worker
from tests.utils import wait_for_controller


@pytest.mark.asyncio
async def test_logs(
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[WorkerName], Awaitable[Worker]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    stream_eiger: Callable[[zmq.Context[Any], int, int], Coroutine[Any, Any, None]],
) -> None:
    await reducer(None)
    await create_worker(WorkerName("w1"))
    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("eiger")],
                upstream_url=Url("tcp://localhost:9999"),
            ),
        )
    )

    await wait_for_controller(streams={StreamName("eiger")})
    async with aiohttp.ClientSession() as session:
        logs = await session.get("http://localhost:5000/api/v1/logs?level=warning")
        lgs = await logs.json()
        # assert len(lgs) > 0 currently there is nothing written to redis from a test, but it works when called from cli
        for log in lgs:
            assert log["levelname"] == "WARNING"



import asyncio
import json
from typing import Awaitable, Callable, Any, Coroutine, Optional

import aiohttp

import pytest
import zmq.asyncio
import zmq
from pydantic_core import Url

from dranspose.ingester import Ingester
from dranspose.ingesters.zmqpull_single import (
    ZmqPullSingleIngester,
    ZmqPullSingleSettings,
)
from dranspose.protocol import (
    RedisKeys,
    StreamName,
    WorkerName,
    VirtualWorker,
    VirtualConstraint,
)

import redis.asyncio as redis

from dranspose.worker import Worker

from tests.utils import wait_for_controller, wait_for_finish


@pytest.mark.asyncio
async def test_simple(
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[WorkerName], Awaitable[Worker]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    stream_eiger: Callable[[zmq.Context[Any], int, int], Coroutine[Any, Any, None]],
) -> None:
    await reducer(None)
    await create_worker(WorkerName("w1"))
    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("eiger")],
                upstream_url=Url("tcp://localhost:9999"),
            ),
        )
    )

    r = redis.Redis(host="localhost", port=6379, decode_responses=True, protocol=3)

    async with aiohttp.ClientSession() as session:
        await wait_for_controller(
            streams={StreamName("eiger")}, workers={WorkerName("w1")}
        )

        ntrig = 10
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json={
                "eiger": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(2 * i)).model_dump(
                            mode="json"
                        )
                    ]
                    for i in range(1, ntrig)
                ],
            },
        )
        assert resp.status == 200
        uuid = await resp.json()

    updates = await r.xread({RedisKeys.updates(): 0})
    assert "dranspose:controller:updates" in updates
    assert json.loads(updates["dranspose:controller:updates"][-1][-1][1]["data"]) == {
        "mapping_uuid": uuid,
        "parameters_version": {},
        "target_parameters_hash": None,
        "active_streams": ["eiger"],
        "finished": False,
    }
    keys = await r.keys("dranspose:*")
    present_keys = {
        "dranspose:controller:updates",
        "dranspose:worker:w1:config",
        f"dranspose:assigned:{uuid}",
        f"dranspose:ready:{uuid}",
        "dranspose:ingester:eiger-ingester:config",
        "dranspose:controller_lock",
        "dranspose:reducer:reducer:config",
    }
    assert present_keys.issubset(keys)

    context = zmq.asyncio.Context()

    asyncio.create_task(stream_eiger(context, 9999, ntrig - 1))

    content = await wait_for_finish()

    assert content == {
        "last_assigned": ntrig + 1,
        "completed_events": ntrig + 1,
        "total_events": ntrig + 1,
        "finished": True,
    }
    context.destroy()

    await r.aclose()


@pytest.mark.asyncio
async def test_map(
    controller: None,
    reducer: Callable[[Optional[str]], Awaitable[None]],
    create_worker: Callable[[WorkerName], Awaitable[Worker]],
    create_ingester: Callable[[Ingester], Awaitable[Ingester]],
    stream_eiger: Callable[[zmq.Context[Any], int, int], Coroutine[Any, Any, None]],
    stream_orca: Callable[[zmq.Context[Any], int, int], Coroutine[Any, Any, None]],
    stream_small: Callable[[zmq.Context[Any], int, int], Coroutine[Any, Any, None]],
) -> None:
    await reducer(None)
    await create_worker(WorkerName("w1"))
    await create_worker(WorkerName("w2"))
    await create_worker(WorkerName("w3"))
    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("eiger")],
                upstream_url=Url("tcp://localhost:9999"),
            ),
        )
    )
    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("orca")],
                upstream_url=Url("tcp://localhost:9998"),
                ingester_url=Url("tcp://localhost:10011"),
            ),
        )
    )
    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("alba")],
                upstream_url=Url("tcp://localhost:9997"),
                ingester_url=Url("tcp://localhost:10012"),
            ),
        )
    )
    await create_ingester(
        ZmqPullSingleIngester(
            settings=ZmqPullSingleSettings(
                ingester_streams=[StreamName("slow")],
                upstream_url=Url("tcp://localhost:9996"),
                ingester_url=Url("tcp://localhost:10013"),
            ),
        )
    )

    await wait_for_controller(
        streams={
            StreamName("eiger"),
            StreamName("orca"),
            StreamName("alba"),
            StreamName("slow"),
        }
    )
    async with aiohttp.ClientSession() as session:
        print("startup done")
        ntrig = 10
        resp = await session.post(
            "http://localhost:5000/api/v1/mapping",
            json={
                "eiger": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(2 * i)).model_dump(
                            mode="json"
                        )
                    ]
                    for i in range(1, ntrig)
                ],
                "orca": [
                    [
                        VirtualWorker(
                            constraint=VirtualConstraint(2 * i + 1)
                        ).model_dump(mode="json")
                    ]
                    for i in range(1, ntrig)
                ],
                "alba": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(2 * i)).model_dump(
                            mode="json"
                        ),
                        VirtualWorker(
                            constraint=VirtualConstraint(2 * i + 1)
                        ).model_dump(mode="json"),
                    ]
                    for i in range(1, ntrig)
                ],
                "slow": [
                    [
                        VirtualWorker(constraint=VirtualConstraint(2 * i)).model_dump(
                            mode="json"
                        ),
                        VirtualWorker(
                            constraint=VirtualConstraint(2 * i + 1)
                        ).model_dump(mode="json"),
                    ]
                    if i % 4 == 0
                    else None
                    for i in range(1, ntrig)
                ],
            },
        )
        assert resp.status == 200

    context = zmq.asyncio.Context()

    asyncio.create_task(stream_eiger(context, 9999, ntrig - 1))
    asyncio.create_task(stream_orca(context, 9998, ntrig - 1))
    asyncio.create_task(stream_small(context, 9997, ntrig - 1))
    asyncio.create_task(stream_small(context, 9996, ntrig // 4))

    content = await wait_for_finish()
    assert content == {
        "last_assigned": ntrig + 1,
        "completed_events": ntrig + 1,
        "total_events": ntrig + 1,
        "finished": True,
    }

    context.destroy()



from typing import AsyncGenerator, Optional

import zmq

from dranspose.data.eiger_legacy import (
    EigerLegacyPacket,
    EigerLegacyHeader,
    EigerLegacyImage,
    EigerLegacyEnd,
)
from dranspose.event import StreamData
from dranspose.ingester import Ingester, IngesterSettings
from dranspose.protocol import StreamName, ZmqUrl


class ZmqPullEigerLegacySettings(IngesterSettings):
    upstream_url: ZmqUrl


class ZmqPullEigerLegacyIngester(Ingester):
    """
    A simple ingester class to comsume a stream from the streaming-receiver repub port
    """

    def __init__(self, settings: Optional[ZmqPullEigerLegacySettings] = None) -> None:
        if settings is not None:
            self._streaming_eiger_legacy_settings = settings
        else:
            self._streaming_eiger_legacy_settings = ZmqPullEigerLegacySettings()

        super().__init__(settings=self._streaming_eiger_legacy_settings)
        self.in_socket: Optional[zmq._future._AsyncSocket] = None

    async def run_source(self, stream: StreamName) -> AsyncGenerator[StreamData, None]:
        self.in_socket = self.ctx.socket(zmq.PULL)
        self.in_socket.connect(str(self._streaming_eiger_legacy_settings.upstream_url))
        self._logger.info(
            "pulling from %s", self._streaming_eiger_legacy_settings.upstream_url
        )

        while True:
            self._logger.debug("clear up insocket")
            parts = await self.in_socket.recv_multipart(copy=False)
            try:
                packet = EigerLegacyPacket.validate_json(parts[0].bytes)
            except Exception as e:
                self._logger.error("packet not valid %s", e.__repr__())
                continue
            self._logger.debug("received frame with header %s", packet)
            if isinstance(packet, EigerLegacyHeader):
                self._logger.info("start of new sequence %s", packet)
                yield StreamData(typ="EIGER_LEGACY", frames=parts)
                break
        while True:
            parts = await self.in_socket.recv_multipart(copy=False)
            try:
                packet = EigerLegacyPacket.validate_json(parts[0].bytes)
            except Exception as e:
                self._logger.error("packet not valid %s", e.__repr__())
                continue
            if isinstance(packet, EigerLegacyImage):
                yield StreamData(typ="EIGER_LEGACY", frames=parts)
            elif isinstance(packet, EigerLegacyEnd):
                yield StreamData(typ="EIGER_LEGACY", frames=parts)
                break
        while True:
            self._logger.debug("discarding messages until next run")
            await self.in_socket.recv_multipart(copy=False)

    async def stop_source(self, stream: StreamName) -> None:
        if self.in_socket:
            self._logger.info("closing socket without linger")
            self.in_socket.close(linger=0)
            self.in_socket = None



import json
from types import UnionType

import zmq

from dranspose.data.pcap import PCAPPacket
from dranspose.event import StreamData


def parse(data: StreamData) -> UnionType:
    """
    Parses a pcap packet

    Arguments:
        data: a frame comming from the pcap tango device

    Returns:
        a PCAPPacket
    """
    assert data.typ == "PCAP"
    assert data.length == 1
    frame = data.frames[0]
    if isinstance(frame, zmq.Frame):
        val = json.loads(frame.bytes)
    else:
        val = json.loads(frame)

    return PCAPPacket.validate_python(val)



from typing import Any

from dranspose.parameters import (
    ParameterBase,
    StrParameter,
    BinaryParameter,
    IntParameter,
    FloatParameter,
    BoolParameter,
)
from dranspose.protocol import ParameterName


class ParamWorker:
    def __init__(self, **kwargs: Any) -> None:
        self.number = 0

    @staticmethod
    def describe_parameters() -> list[ParameterBase]:
        params = [
            StrParameter(name=ParameterName("roi1"), default="bla"),
            StrParameter(name=ParameterName("str_param")),
            BinaryParameter(name=ParameterName("bytes_param")),
            IntParameter(name=ParameterName("int_param")),
            FloatParameter(name=ParameterName("float_param")),
            BoolParameter(name=ParameterName("bool_param")),
        ]
        return params

    def process_event(self, event, parameters, *args, **kwargs):
        return parameters



from typing import Literal

import zmq
from pydantic import BaseModel, TypeAdapter, ConfigDict

from dranspose.event import StreamData


class Stream1(BaseModel):
    msg_number: int

    def to_stream_data(self) -> StreamData:
        js = self.model_dump_json(exclude={"data"}).encode()
        frames = [zmq.Frame(js)]
        if hasattr(self, "data"):
            frames.append(zmq.Frame(self.data.tobytes()))
        return StreamData(typ="STINS", frames=frames)


class Stream1Start(Stream1):
    """
    Example:
        ``` py
        Stream1Start(
            msg_number=374,
            htype='header',
            filename='/data/visitors/....test.h5'
        )
        ```
    """

    model_config = ConfigDict(extra="allow")

    htype: Literal["header"] = "header"
    filename: str


class Stream1Data(Stream1):
    """
    Example:
        ``` py
        Stream1Data(
            msg_number=375,
            htype='image',
            frame=0,
            shape=[831, 1475],
            type='float32',
            compression='none',
            data=array([[0., 0., 0., ..., 0., 0., 0.],
                [0., 0., 0., ..., 0., 0., 0.],
                [0., 0., 0., ..., 0., 0., 0.],
                ...,
                [0., 0., 0., ..., 0., 0., 0.],
                [0., 0., 0., ..., 0., 0., 0.],
                [0., 0., 0., ..., 0., 0., 0.]], dtype=float32)
        )
        ```
    """

    model_config = ConfigDict(extra="allow")

    htype: Literal["image"] = "image"
    frame: int
    shape: list[int]
    type: str
    compression: str


class Stream1End(Stream1):
    """
    Example:
        ``` py
        Stream1End(
            msg_number=376,
            htype='series_end'
        )
        ```
    """

    model_config = ConfigDict(extra="allow")

    htype: Literal["series_end"] = "series_end"


Stream1Packet = TypeAdapter(Stream1Start | Stream1Data | Stream1End)
"""
A union type for STINS packets
"""



from dataclasses import dataclass
import logging
from typing import Any

import numpy as np

from dranspose.data.lecroy import LecroyParsed
from dranspose.event import EventData, ResultData
from dranspose.middlewares.lecroy import parse
from dranspose.parameters import IntParameter
from dranspose.protocol import ParameterName, StreamName, WorkParameter

logger = logging.getLogger(__name__)


@dataclass
class Result:
    max_val: np.typing.NDArray[Any]
    ts: list[float]


class LecroyWorker:
    def __init__(self, *args: tuple[Any], **kwargs: dict[str, Any]) -> None:
        logger.info("LecroyWorker init")

    @staticmethod
    def describe_parameters() -> list[IntParameter]:
        params = [
            IntParameter(name=ParameterName("channel"), default=2),
        ]
        return params

    def process_event(
        self,
        event: EventData,
        parameters: dict[ParameterName, WorkParameter],
        *args: tuple[Any],
        **kwargs: dict[str, Any],
    ) -> Result | None:
        logger.debug(f"LecroyWorker {event=}")
        if "lecroy" in event.streams:
            res = parse(event.streams[StreamName("lecroy")])
            if isinstance(res, LecroyParsed):
                for i, ch in enumerate(res.channels):
                    logger.info(f"looking at channel {ch}")
                    logger.info(f"meta packet is {res.meta[i]}")
                    logger.info(f"trace shape is {res.data[i].dtype}")
                    logger.info(f"timestamps list len is {len(res.timestamps[i])}")
                    logger.debug(f"traces {res.data[i]}")
                    logger.debug(f"timestamps list {res.timestamps[i]}")
                    if ch == parameters[ParameterName("channel")].value:
                        return Result(np.max(res.data[i], axis=1), res.timestamps[i])
            else:
                logger.info("control packet is %s", res)
        return None


class LecroyReducer:
    def __init__(
        self,
        parameters: dict[ParameterName, WorkParameter] | None = None,
        *args: tuple[Any],
        **kwargs: dict[str, Any],
    ) -> None:
        self.publish: dict[str, Any] = {"max_val": [], "ts": []}

    def process_result(
        self, result: ResultData, parameters: dict[ParameterName, WorkParameter]
    ) -> None:
        if result.payload and isinstance(result.payload, Result):
            logger.info(f"max_val is {result.payload.max_val}")
            logger.info(f"timestamps is {result.payload.ts}")
            self.publish["max_val"] = result.payload.max_val
            self.publish["ts"] = result.payload.ts



class FluorescenceWorker:
    @staticmethod
    def describe_parameters():
        return [IntParameter(name="roi1-start")]

    def __init__(self, parameters=None):
        self.number = 0

    def process_event(self, event: EventData, parameters=None):
        print(event)
        # parse zmq frames, fit spectra to get concentrations, extract motor position
        return {"position": mot, "concentrations": ...}



from typing import Literal
from datetime import datetime

import zmq
from pydantic import BaseModel, TypeAdapter, ConfigDict

from dranspose.event import StreamData


class PCAPBase(BaseModel):
    message_id: int
    version: Literal[1]

    def to_stream_data(self) -> StreamData:
        dat = self.model_dump_json()
        return StreamData(typ="PCAP", frames=[zmq.Frame(dat)])


class PCAPStart(PCAPBase):
    """
    Example:
        ``` py
        PCAPStart(
            message_id=1,
            version=1,
            message_type='series-start'
            arm_time="2024-10-07T08:49:10.627Z"
        )
        ```
    """

    model_config = ConfigDict(extra="allow")

    message_type: Literal["series-start"]
    arm_time: datetime


class PCAPData(PCAPBase):
    """
    While the original stream sends 3 separate zmq frames (no multipart), this returns a single packet.

    Example:
        ``` py
        PCAPImage(
            message_id=2,
            version=1,
            message_type="data",
            frame_number=0,
            inttime=[7.1160000000000005, 7.315, 7.093800000000001, 7.1584, 7.1322, 7.1062, 7.0206, 6.9082, 7.0164, 6.912, 6.966, 6.885000000000001, 6.7946, 6.731800000000001, 6.9636000000000005],
            triggernumber=[1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0, 14.0, 15.0],
            repeatindex=[1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
            energy=[8879.101048716273, 8879.299376689953, 8879.5003861779, 8879.699330779304, 8879.8999612314, 8880.099636263549, 8880.300291637586, 8880.499422033765, 8880.700264322526, 8880.900116547104, 8881.099615570443, 8881.29966467182, 8881.499998985424, 8881.700179495409, 8881.900461311128],
            INENC3.VAL.Mean=[0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
        ```
    """

    model_config = ConfigDict(extra="allow")

    message_type: Literal["data"]
    frame_number: int


class PCAPEnd(PCAPBase):
    """
    Example:
        ``` py
        PCAPEnd(
            message_id=6,
            version=1,
            message_type='series-end')
        ```
    """

    model_config = ConfigDict(extra="allow")

    message_type: Literal["series-end"]


PCAPPacket = TypeAdapter(PCAPStart | PCAPData | PCAPEnd)
"""
Union type for PCAP packets
"""



import logging
import pytest

from dranspose.mapping import NotYetAssigned, MappingSequence, Map
from dranspose.protocol import (
    VirtualWorker,
    StreamName,
    WorkerName,
    EventNumber,
    WorkAssignment,
    WorkerState,
    VirtualConstraint,
    WorkerTag,
    MappingName,
)


def test_expand() -> None:
    ntrig = 10
    p = MappingSequence(
        parts={
            MappingName("main"): {
                StreamName("test"): [
                    [VirtualWorker(constraint=VirtualConstraint(i))]
                    for i in range(ntrig)
                ]
            }
        },
        sequence=[MappingName("main")],
        add_start_end=False,
    )
    mapping = p.expand()
    assert mapping == {
        StreamName("test"): [
            [VirtualWorker(constraint=VirtualConstraint(i))] for i in range(ntrig)
        ]
    }


def test_expand_double() -> None:
    ntrig = 10
    p = MappingSequence(
        parts={
            MappingName("main"): {
                StreamName("test"): [
                    [VirtualWorker(constraint=VirtualConstraint(i))]
                    for i in range(ntrig)
                ]
            }
        },
        sequence=[MappingName("main"), MappingName("main")],
        add_start_end=False,
    )
    mapping = p.expand()
    assert mapping == {
        StreamName("test"): [
            [VirtualWorker(constraint=VirtualConstraint(i))] for i in range(ntrig)
        ]
        + [[VirtualWorker(constraint=VirtualConstraint(i))] for i in range(ntrig)]
    }


def test_expand_double_streams() -> None:
    ntrig = 10
    p = MappingSequence(
        parts={
            MappingName("main"): {
                StreamName("test"): [
                    [VirtualWorker(constraint=VirtualConstraint(i))]
                    for i in range(ntrig)
                ],
                StreamName("two"): [
                    [VirtualWorker(constraint=VirtualConstraint(i))]
                    for i in range(ntrig)
                ],
            },
            MappingName("two"): {
                StreamName("two"): [
                    [VirtualWorker(constraint=VirtualConstraint(i))]
                    for i in range(ntrig)
                ]
            },
        },
        sequence=[MappingName("two"), MappingName("main")],
        add_start_end=False,
    )
    mapping = p.expand()
    assert mapping == {
        StreamName("test"): [None for i in range(ntrig)]
        + [[VirtualWorker(constraint=VirtualConstraint(i))] for i in range(ntrig)],
        StreamName("two"): [
            [VirtualWorker(constraint=VirtualConstraint(i))] for i in range(ntrig)
        ]
        + [[VirtualWorker(constraint=VirtualConstraint(i))] for i in range(ntrig)],
    }
    assert p.len() == 2 * ntrig


def test_wrong_sequence() -> None:
    with pytest.raises(AssertionError):
        ntrig = 6
        MappingSequence(
            parts={
                MappingName("main"): {
                    StreamName("test"): [
                        [VirtualWorker(constraint=VirtualConstraint(i))]
                        for i in range(ntrig)
                    ]
                }
            },
            sequence=[MappingName("two")],
            add_start_end=False,
        )


def test_huge() -> None:
    ntrig = 1000
    reps = 980
    p = MappingSequence(
        parts={
            MappingName("main"): {
                StreamName("test"): [
                    [VirtualWorker(constraint=VirtualConstraint(i))]
                    for i in range(ntrig)
                ]
            }
        },
        sequence=[MappingName("main")] * reps,
        add_start_end=False,
    )
    assert p.len() == ntrig * reps


def test_useless_worker() -> None:
    ntrig = 10
    reps = 20
    m = MappingSequence(
        parts={
            MappingName("main"): {
                StreamName("test"): [
                    [VirtualWorker(constraint=VirtualConstraint(i))]
                    for i in range(ntrig)
                ]
            },
            MappingName("next"): {
                StreamName("test"): [
                    [
                        VirtualWorker(
                            constraint=VirtualConstraint(i), tags={WorkerTag("useless")}
                        )
                    ]
                    for i in range(ntrig)
                ]
            },
        },
        sequence=[MappingName("main")] * reps + [MappingName("next")] * reps,
        add_start_end=False,
    )
    assert m.len() == ntrig * reps * 2

    all_workers = [
        WorkerState(
            name=WorkerName("useless"),
            tags={WorkerTag("useless")},
        ),
        WorkerState(name=WorkerName("w1")),
    ]

    ret = m.assign_next(all_workers[0], all_workers, future=100)
    assert ret == []

    assert len(m.active_maps) == 12
    for i in range(200):
        ret = m.assign_next(all_workers[1], all_workers, future=100)
        assert len(ret) == 1, f"failed assign {i}"

    assert len(m.active_maps) == 21

    for i in range(199):
        ret = m.assign_next(all_workers[0], all_workers, future=100)
        assert len(ret) == 1, f"failed assign {i}"

    for i in range(400):
        workers = m.get_event_workers(EventNumber(i))
        if i < 200:
            assert workers == WorkAssignment(
                event_number=EventNumber(i),
                assignments={StreamName("test"): [WorkerName("w1")]},
            )
        else:
            assert workers == WorkAssignment(
                event_number=EventNumber(i),
                assignments={StreamName("test"): [WorkerName("useless")]},
            )


def test_none() -> None:
    m = MappingSequence(
        parts={
            MappingName("main"): {
                StreamName("test"): [
                    [VirtualWorker(constraint=VirtualConstraint(i))]
                    if i % 4 == 0
                    else None
                    for i in range(4)
                ]
            },
            MappingName("end"): {
                StreamName("test"): [
                    [VirtualWorker(constraint=VirtualConstraint(i))]
                    if i % 4 == 0
                    else None
                    for i in range(2)
                ]
            },
        },
        sequence=[MappingName("main"), MappingName("main"), MappingName("end")],
        add_start_end=False,
    )

    m.print()
    all_workers = [
        WorkerState(name=WorkerName("w1")),
        WorkerState(name=WorkerName("w2")),
    ]
    m.assign_next(all_workers[0], all_workers)
    m.assign_next(all_workers[1], all_workers)

    m.print()

    with pytest.raises(NotYetAssigned):
        m.get_event_workers(EventNumber(8))

    logging.info("assign last one to have ev 8")
    ret = m.assign_next(all_workers[0], all_workers)
    assert ret == [VirtualWorker(constraint=VirtualConstraint(0))]
    logging.info("assigned last worker to %s", ret)

    m.print()
    assign = m.get_event_workers(EventNumber(8))
    assert assign.assignments[StreamName("test")] == [WorkerName("w1")]


def test_discard() -> None:
    ntrig = 10
    m = MappingSequence(
        parts={
            MappingName("main"): {StreamName("test"): [[] for i in range(ntrig // 2)]}
        },
        sequence=[MappingName("main"), MappingName("main")],
        add_start_end=False,
    )

    m.print()
    all_workers = [
        WorkerState(name=WorkerName("w1")),
        WorkerState(name=WorkerName("w2")),
    ]
    r = m.assign_next(all_workers[0], all_workers)
    print(r)
    r = m.assign_next(all_workers[1], all_workers)
    print(r)

    m.print()
    print("completed", m.complete_events)

    # m.assign_next(all_workers[0], all_workers)

    assign = m.get_event_workers(EventNumber(8))
    print("assign is", assign)
    assert assign.assignments == {"test": []}


def test_partial_discard() -> None:
    m = MappingSequence(
        parts={
            MappingName("use"): {
                StreamName("test"): [
                    [VirtualWorker(constraint=VirtualConstraint(i))] for i in range(2)
                ]
            },
            MappingName("discard"): {StreamName("test"): [[] for i in range(3)]},
            MappingName("none"): {StreamName("test"): [None for i in range(3)]},
        },
        sequence=[
            MappingName("use"),
            MappingName("discard"),
            MappingName("none"),
            MappingName("use"),
        ],
        add_start_end=False,
    )

    m.print()
    all_workers = [
        WorkerState(name=WorkerName("w1")),
        WorkerState(name=WorkerName("w2")),
    ]
    r = m.assign_next(all_workers[0], all_workers)
    print(r)
    r = m.assign_next(all_workers[1], all_workers)
    print(r)
    r = m.assign_next(all_workers[0], all_workers)
    print(r)

    m.print()
    print("completed", m.complete_events)

    # m.assign_next(all_workers[0], all_workers)

    assign = m.get_event_workers(EventNumber(4))
    assert assign.assignments == {"test": []}
    assign = m.get_event_workers(EventNumber(6))
    assert assign.assignments == {}
    assign = m.get_event_workers(EventNumber(8))
    assert assign.assignments == {"test": ["w1"]}


def test_discard_only_one() -> None:
    ntrig = 10
    m = MappingSequence(
        parts={
            MappingName("main"): {
                StreamName("test"): [[] for i in range(ntrig // 2)],
                StreamName("test2"): [
                    [VirtualWorker(constraint=VirtualConstraint(i))]
                    for i in range(ntrig // 2)
                ],
            }
        },
        sequence=[MappingName("main"), MappingName("main")],
        add_start_end=False,
    )

    m.print()
    all_workers = [
        WorkerState(name=WorkerName("w1")),
        WorkerState(name=WorkerName("w2")),
    ]
    m.assign_next(all_workers[0], all_workers)
    m.assign_next(all_workers[1], all_workers)

    m.print()
    print("completed", m.complete_events)

    with pytest.raises(NotYetAssigned):
        m.get_event_workers(EventNumber(8))

    m.assign_next(all_workers[0], all_workers)

    assign = m.get_event_workers(EventNumber(2))
    assert assign.assignments[StreamName("test2")] == [WorkerName("w1")]


def test_auto() -> None:
    m = MappingSequence(
        parts={
            MappingName("main"): {
                StreamName("test"): [
                    [VirtualWorker(constraint=VirtualConstraint(i))]
                    if i % 4 == 0
                    else None
                    for i in range(4)
                ]
            },
            MappingName("end"): {
                StreamName("test"): [
                    [VirtualWorker(constraint=VirtualConstraint(i))]
                    if i % 4 == 0
                    else None
                    for i in range(2)
                ]
            },
        },
        sequence=[MappingName("main"), MappingName("main"), MappingName("end")],
    )

    m.print()
    all_workers = [
        WorkerState(name=WorkerName("w1")),
        WorkerState(name=WorkerName("w2")),
    ]
    m.assign_next(all_workers[0], all_workers)
    m.assign_next(all_workers[1], all_workers)

    m.print()

    assert 1 == m.complete_events

    with pytest.raises(NotYetAssigned):
        m.get_event_workers(EventNumber(8))


def test_all() -> None:
    m = MappingSequence(
        parts={
            MappingName("main"): {
                StreamName("test"): [
                    [VirtualWorker(constraint=VirtualConstraint(i))]
                    if i % 4
                    else [VirtualWorker()]
                    for i in range(4)
                ]
            },
            MappingName("end"): {
                StreamName("test"): [
                    [VirtualWorker(constraint=VirtualConstraint(i))]
                    if i % 4
                    else [VirtualWorker()]
                    for i in range(2)
                ]
            },
        },
        sequence=[MappingName("main"), MappingName("main"), MappingName("end")],
        add_start_end=False,
    )

    print("before assignment")
    m.print()
    all_workers = [
        WorkerState(name=WorkerName("w1")),
        WorkerState(name=WorkerName("w2")),
    ]
    for _ in range(3):
        m.assign_next(all_workers[0], all_workers)
        m.assign_next(all_workers[1], all_workers)

    print("assigned 3x w1,w2")
    m.print()
    # print("assignments", m.assignments, m.all_assignments)

    with pytest.raises(NotYetAssigned):
        m.get_event_workers(EventNumber(4))

    m.assign_next(all_workers[0], all_workers)

    print("assigned another w1")
    m.print()

    assign = m.get_event_workers(EventNumber(4))
    assert set(assign.assignments[StreamName("test")]) == set(
        [ws.name for ws in all_workers]
    )


def test_multi_all() -> None:
    m = MappingSequence(
        parts={
            MappingName("main"): {
                StreamName("test"): [
                    [VirtualWorker(constraint=VirtualConstraint(i))]
                    if i % 4
                    else [VirtualWorker()]
                    for i in range(4)
                ],
                StreamName("test2"): [
                    [VirtualWorker(constraint=VirtualConstraint(i))]
                    if i % 4
                    else [VirtualWorker()]
                    for i in range(4)
                ],
                StreamName("noall"): [
                    [VirtualWorker(constraint=VirtualConstraint(i))] for i in range(4)
                ],
            },
            MappingName("end"): {
                StreamName("test"): [
                    [VirtualWorker(constraint=VirtualConstraint(i))]
                    if i % 4
                    else [VirtualWorker()]
                    for i in range(2)
                ],
                StreamName("test2"): [
                    [VirtualWorker(constraint=VirtualConstraint(i))]
                    if i % 4
                    else [VirtualWorker()]
                    for i in range(2)
                ],
                StreamName("noall"): [
                    [VirtualWorker(constraint=VirtualConstraint(i))] for i in range(2)
                ],
            },
        },
        sequence=[MappingName("main"), MappingName("main"), MappingName("end")],
        add_start_end=False,
    )

    m.print()
    all_workers = [
        WorkerState(name=WorkerName("w1")),
        WorkerState(name=WorkerName("w2")),
        WorkerState(name=WorkerName("w3")),
    ]
    for _ in range(3):
        m.assign_next(all_workers[0], all_workers)
        m.assign_next(all_workers[1], all_workers)
        m.assign_next(all_workers[2], all_workers)

    m.print()

    m.assign_next(all_workers[0], all_workers)

    m.print()

    assign = m.get_event_workers(EventNumber(4))
    assert set(assign.assignments[StreamName("test")]) == set(
        [ws.name for ws in all_workers]
    )


def test_multiple() -> None:
    m = MappingSequence(
        parts={
            MappingName("main"): {
                StreamName("eiger"): [
                    [VirtualWorker(constraint=VirtualConstraint(2 * i))]
                    for i in range(1, 6)
                ],
                StreamName("alba"): [
                    [
                        VirtualWorker(constraint=VirtualConstraint(2 * i)),
                        VirtualWorker(constraint=VirtualConstraint(2 * i + 1)),
                    ]
                    for i in range(1, 6)
                ],
                StreamName("orca"): [
                    [VirtualWorker(constraint=VirtualConstraint(2 * i + 1))]
                    for i in range(1, 6)
                ],
            }
        },
        sequence=[MappingName("main"), MappingName("main")],
        add_start_end=False,
    )
    m.print()
    all_workers = [
        WorkerState(name=WorkerName("w1")),
        WorkerState(name=WorkerName("w2")),
        WorkerState(name=WorkerName("w3")),
    ]
    for i in range(5):
        print(m.complete_events)
        m.assign_next(all_workers[0], all_workers)
        m.assign_next(all_workers[1], all_workers)
        m.assign_next(all_workers[2], all_workers)
        print(m.complete_events)
        print("--")
        # m.assign_next("w3")
    # print(m.assignments)
    # print(m.all_assignments)
    assert m.complete_events == 7

    evworkers = m.get_event_workers(EventNumber(m.complete_events - 1))
    print(evworkers)
    assert evworkers == WorkAssignment(
        event_number=EventNumber(6),
        assignments={
            StreamName("eiger"): [WorkerName("w1")],
            StreamName("alba"): [WorkerName("w1"), WorkerName("w2")],
            StreamName("orca"): [WorkerName("w2")],
        },
    )
    assert m.min_workers() == 2


def test_mixed_all() -> None:
    m = MappingSequence(
        parts={
            MappingName("main"): {
                StreamName("eiger"): [
                    [VirtualWorker(constraint=VirtualConstraint(2 * i))]
                    for i in range(1, 6)
                ],
                StreamName("orca"): [
                    [VirtualWorker(constraint=VirtualConstraint(2 * i + 1))]
                    for i in range(1, 6)
                ],
                StreamName("announcer"): [[VirtualWorker()] for i in range(1, 6)],
            }
        },
        sequence=[MappingName("main"), MappingName("main")],
        add_start_end=False,
    )
    m.print()

    all_workers = [
        WorkerState(name=WorkerName("w1")),
        WorkerState(name=WorkerName("w2")),
        WorkerState(name=WorkerName("w3")),
    ]
    for i in range(4):
        print(m.complete_events)
        m.assign_next(all_workers[0], all_workers)
        m.assign_next(all_workers[1], all_workers)
        m.assign_next(all_workers[2], all_workers)
        print(m.complete_events)
        print("--")
        # m.assign_next("w3")
    # print(m.assignments)
    # print(m.all_assignments)
    m.print()
    assert m.active_maps[0].all_assignments[
        (EventNumber(1), StreamName("announcer"), 0)
    ] == [
        "w1",
        "w2",
        "w3",
    ]

    m.print()


def test_all_wrap_tags() -> None:
    m = MappingSequence(
        parts={
            MappingName("main"): {
                StreamName("test"): [
                    [VirtualWorker(constraint=VirtualConstraint(i))]
                    if i % 4
                    else [VirtualWorker(tags={WorkerTag("tag"), WorkerTag("bla")})]
                    for i in range(4)
                ],
                StreamName("test2"): [
                    [VirtualWorker(constraint=VirtualConstraint(i))]
                    if i % 4
                    else [VirtualWorker(tags={WorkerTag("tag2")})]
                    for i in range(4)
                ],
            },
            MappingName("end"): {
                StreamName("test"): [
                    [VirtualWorker(constraint=VirtualConstraint(i))]
                    if i % 4
                    else [VirtualWorker(tags={WorkerTag("tag"), WorkerTag("bla")})]
                    for i in range(2)
                ],
                StreamName("test2"): [
                    [VirtualWorker(constraint=VirtualConstraint(i))]
                    if i % 4
                    else [VirtualWorker(tags={WorkerTag("tag2")})]
                    for i in range(2)
                ],
            },
        },
        sequence=[MappingName("main"), MappingName("main"), MappingName("end")],
        add_start_end=True,
    )
    print("before assignment")
    m.print()

    first_test = m.active_maps[0].mapping[StreamName("test")][0]
    assert first_test is not None
    assert sorted([tuple(t.tags) for t in first_test]) == sorted(
        [("generic",), ("tag",), ("bla",), ("tag2",)]
    )

    all_workers = [
        WorkerState(
            name=WorkerName("w1"),
            tags={WorkerTag("generic"), WorkerTag("tag"), WorkerTag("bla")},
        ),
        WorkerState(
            name=WorkerName("w2"), tags={WorkerTag("tag2"), WorkerTag("generic")}
        ),
    ]
    for i in range(4):
        print(m.complete_events)
        m.assign_next(all_workers[0], all_workers)
        m.assign_next(all_workers[1], all_workers)
        print(m.complete_events)
        print("--")
        # m.assign_next("w3")
    # print(m.assignments)
    # print(m.all_assignments)
    m.print()

    evworkers = m.get_event_workers(EventNumber(0))
    print(evworkers)
    should = {"test": {"w2", "w1"}, "test2": {"w2", "w1"}}
    for st, wn in evworkers.assignments.items():
        assert set(wn) == should[st]


def test_uniform() -> None:
    m = Map.from_uniform({StreamName("orca"), StreamName("panda")}, 10)
    s = MappingSequence(
        parts={MappingName("sardana"): m.mapping},
        sequence=[MappingName("sardana")],
        add_start_end=True,
    )
    assert s.len() == 12
    logging.info("map %s", m)
    b = m.mapping[StreamName("orca")][9]
    assert b is not None
    a = b[0]
    assert isinstance(a, VirtualWorker)
    assert a.constraint == 9


def test_huge_map() -> None:
    ntrig = 10_000
    m = MappingSequence(
        parts={
            MappingName("main"): {
                StreamName("test"): [
                    [VirtualWorker(constraint=VirtualConstraint(i))]
                    for i in range(ntrig)
                ]
            }
        },
        sequence=[MappingName("main")],
        add_start_end=False,
    )

    assert m.len() == ntrig



import json
from types import UnionType

import zmq

from dranspose.data.sardana import SardanaPacket
from dranspose.event import StreamData


def parse(data: StreamData) -> UnionType:
    """
    Parses a sardana packet, which returns a dict

    Arguments:
        data: a frame comming from sardana

    Returns:
        a SardanaPacket
    """
    assert data.typ == "sardana"
    assert data.length == 1
    frame = data.frames[0]
    if isinstance(frame, zmq.Frame):
        val = json.loads(frame.bytes)
    else:
        val = json.loads(frame)

    return SardanaPacket.validate_python(val)



import pickle
from typing import Literal, Optional

import zmq
from pydantic import BaseModel, TypeAdapter, ConfigDict

from dranspose.event import StreamData


class XspressBase(BaseModel):
    def to_stream_data(self) -> StreamData:
        js = self.model_dump_json(exclude={"data", "meta"}).encode()
        frames = [zmq.Frame(js)]
        if hasattr(self, "data"):
            frames.append(zmq.Frame(self.data.tobytes()))
        if hasattr(self, "meta"):
            pk = pickle.dumps(list(self.meta.values()))
            frames.append(zmq.Frame(pk))
        return StreamData(typ="xspress", frames=frames)


class XspressStart(XspressBase):
    """
    Example:
        ``` py
        XspressStart(
            htype='header',
            filename='/data/.../diff_1130_stream_test/raw/dummy/scan_000002_xspress3.hdf5',
            overwritable=False
        )
        ```
    """

    model_config = ConfigDict(extra="allow")

    htype: Literal["header"] = "header"
    filename: str


class XspressImage(XspressBase):
    """
    While the original stream sends 3 separate zmq frames (no multipart), this returns a single packet.

    Example:
        ``` py
        XspressImage(
            htype='image',
            frame=0,
            shape=[4, 4096],
            exptime=0.099999875,
            type='uint32',
            compression='none',
            data=array([[0, 0, 0, ..., 0, 0, 0],
               [0, 0, 0, ..., 0, 0, 0],
               [0, 0, 0, ..., 0, 0, 0],
               [0, 0, 0, ..., 0, 0, 0]], dtype=uint32),
            meta={
                'ocr': array([0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 1.56250195e-15]),
                'AllEvents': array([2, 0, 0, 3], dtype=uint32),
                'AllGood': array([0, 0, 0, 1], dtype=uint32),
                'ClockTicks': array([7999990, 7999990, 7999990, 7999990], dtype=uint32),
                'TotalTicks': array([7999990, 7999990, 7999990, 7999990], dtype=uint32),
                'ResetTicks': array([ 0,  0,  0, 91], dtype=uint32),
                'event_widths': array([6, 6, 6, 6], dtype=int32),
                'dtc': array([1.00000175, 1.        , 1.        , 1.000014  ])
            }
        )
        ```
    """

    model_config = ConfigDict(extra="allow")

    htype: Literal["image"] = "image"
    frame: int
    shape: list[int]
    exptime: Optional[float] = 1
    type: str
    compression: Optional[str] = "none"


class XspressEnd(XspressBase):
    """
    Example:
        ``` py
        XspressEnd(htype='series_end')
        ```
    """

    model_config = ConfigDict(extra="allow")

    htype: Literal["series_end"] = "series_end"


XspressPacket = TypeAdapter(XspressStart | XspressImage | XspressEnd)
"""
Union type for Xspress packets
"""
